"""Ludwig string tokenizers including string-based, spacy-based, and huggingface-based implementations.

To add a new tokenizer, 1) implement a subclass of BaseTokenizer and 2) add it to the tokenizer_registry.

Once it's in the registry, tokenizers can be used in a ludwig config, e.g..

```
input_features:
    -   name: title
        type: text
        preprocessing:
            tokenizer: <NEW_TOKENIZER>
```
"""

import logging
import re
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Union

import torch

from ludwig.constants import PADDING_SYMBOL, UNKNOWN_SYMBOL
from ludwig.utils.data_utils import load_json
from ludwig.utils.hf_utils import load_pretrained_hf_tokenizer
from ludwig.utils.nlp_utils import load_nlp_pipeline, process_text

logger = logging.getLogger(__name__)

SPACE_PUNCTUATION_REGEX = re.compile(r"\w+|[^\w\s]")
COMMA_REGEX = re.compile(r"\s*,\s*")
UNDERSCORE_REGEX = re.compile(r"\s*_\s*")

TORCHSCRIPT_COMPATIBLE_TOKENIZERS = {"space", "space_punct"}
TORCHTEXT_0_12_0_TOKENIZERS = {"sentencepiece", "clip", "gpt2bpe"}
TORCHTEXT_0_13_0_TOKENIZERS = {"bert"}

HF_TOKENIZER_SAMPLE_INPUTS = ["UNwant\u00E9d,running", "ah\u535A\u63A8zz", " \tHeLLo!how  \n Are yoU? [UNK]"]


class BaseTokenizer:
    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def __call__(self, text: str):
        pass


class CharactersToListTokenizer(BaseTokenizer):
    def __call__(self, text):
        return [char for char in text]


class SpaceStringToListTokenizer(torch.nn.Module):
    """Implements torchscript-compatible whitespace tokenization."""

    def __init__(self, **kwargs):
        super().__init__()

    def forward(self, v: Union[str, List[str], torch.Tensor]) -> Any:
        if isinstance(v, torch.Tensor):
            raise ValueError(f"Unsupported input: {v}")

        inputs: List[str] = []
        # Ludwig calls map on List[str] objects, so we need to handle individual strings as well.
        if isinstance(v, str):
            inputs.append(v)
        else:
            inputs.extend(v)

        tokens: List[List[str]] = []
        for sequence in inputs:
            split_sequence = sequence.strip().split(" ")
            token_sequence: List[str] = []
            for token in self.get_tokens(split_sequence):
                if len(token) > 0:
                    token_sequence.append(token)
            tokens.append(token_sequence)

        return tokens[0] if isinstance(v, str) else tokens

    def get_tokens(self, tokens: List[str]) -> List[str]:
        return tokens


class NgramTokenizer(SpaceStringToListTokenizer):
    """Implements torchscript-compatible n-gram tokenization."""

    def __init__(self, ngram_size: int = 2, **kwargs):
        super().__init__()
        self.n = ngram_size or 2

    def get_tokens(self, tokens: List[str]) -> List[str]:
        from torchtext.data.utils import ngrams_iterator

        return list(ngrams_iterator(tokens, ngrams=self.n))


class SpacePunctuationStringToListTokenizer(torch.nn.Module):
    """Implements torchscript-compatible space_punct tokenization."""

    def __init__(self, **kwargs):
        super().__init__()

    def is_regex_w(self, c: str) -> bool:
        return c.isalnum() or c == "_"

    def forward(self, v: Union[str, List[str], torch.Tensor]) -> Any:
        if isinstance(v, torch.Tensor):
            raise ValueError(f"Unsupported input: {v}")

        inputs: List[str] = []
        # Ludwig calls map on List[str] objects, so we need to handle individual strings as well.
        if isinstance(v, str):
            inputs.append(v)
        else:
            inputs.extend(v)

        tokens: List[List[str]] = []
        for sequence in inputs:
            token_sequence: List[str] = []
            word: List[str] = []
            for c in sequence:
                if self.is_regex_w(c):
                    word.append(c)
                elif len(word) > 0:  # if non-empty word and non-alphanumeric char, append word to token sequence
                    token_sequence.append("".join(word))
                    word.clear()

                if not self.is_regex_w(c) and not c.isspace():  # non-alphanumeric, non-space char is punctuation
                    token_sequence.append(c)

            if len(word) > 0:  # add last word
                token_sequence.append("".join(word))

            tokens.append(token_sequence)

        return tokens[0] if isinstance(v, str) else tokens


class UnderscoreStringToListTokenizer(BaseTokenizer):
    def __call__(self, text):
        return UNDERSCORE_REGEX.split(text.strip())


class CommaStringToListTokenizer(BaseTokenizer):
    def __call__(self, text):
        return COMMA_REGEX.split(text.strip())


class UntokenizedStringToListTokenizer(BaseTokenizer):
    def __call__(self, text):
        return [text]


class StrippedStringToListTokenizer(BaseTokenizer):
    def __call__(self, text):
        return [text.strip()]


class EnglishTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("en"))


class EnglishFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("en"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class EnglishRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("en"), filter_stopwords=True)


class EnglishLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        process_text(text, load_nlp_pipeline("en"), return_lemma=True)


class EnglishLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("en"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class EnglishLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("en"), return_lemma=True, filter_stopwords=True)


class ItalianTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("it"))


class ItalianFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("it"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class ItalianRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("it"), filter_stopwords=True)


class ItalianLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("it"), return_lemma=True)


class ItalianLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("it"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class ItalianLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("it"), return_lemma=True, filter_stopwords=True)


class SpanishTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("es"))


class SpanishFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("es"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class SpanishRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("es"), filter_stopwords=True)


class SpanishLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("es"), return_lemma=True)


class SpanishLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("es"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class SpanishLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("es"), return_lemma=True, filter_stopwords=True)


class GermanTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("de"))


class GermanFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("de"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class GermanRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("de"), filter_stopwords=True)


class GermanLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("de"), return_lemma=True)


class GermanLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("de"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class GermanLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("de"), return_lemma=True, filter_stopwords=True)


class FrenchTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("fr"))


class FrenchFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("fr"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class FrenchRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("fr"), filter_stopwords=True)


class FrenchLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("fr"), return_lemma=True)


class FrenchLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("fr"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class FrenchLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("fr"), return_lemma=True, filter_stopwords=True)


class PortugueseTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pt"))


class PortugueseFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("pt"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class PortugueseRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pt"), filter_stopwords=True)


class PortugueseLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pt"), return_lemma=True)


class PortugueseLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("pt"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class PortugueseLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pt"), return_lemma=True, filter_stopwords=True)


class DutchTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nl"))


class DutchFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("nl"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class DutchRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nl"), filter_stopwords=True)


class DutchLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nl"), return_lemma=True)


class DutchLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("nl"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class DutchLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nl"), return_lemma=True, filter_stopwords=True)


class GreekTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("el"))


class GreekFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("el"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class GreekRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("el"), filter_stopwords=True)


class GreekLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("el"), return_lemma=True)


class GreekLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("el"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class GreekLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("el"), return_lemma=True, filter_stopwords=True)


class NorwegianTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nb"))


class NorwegianFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("nb"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class NorwegianRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nb"), filter_stopwords=True)


class NorwegianLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nb"), return_lemma=True)


class NorwegianLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("nb"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class NorwegianLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("nb"), return_lemma=True, filter_stopwords=True)


class LithuanianTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("lt"))


class LithuanianFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("lt"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class LithuanianRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("lt"), filter_stopwords=True)


class LithuanianLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("lt"), return_lemma=True)


class LithuanianLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("lt"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class LithuanianLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("lt"), return_lemma=True, filter_stopwords=True)


class DanishTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("da"))


class DanishFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("da"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class DanishRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("da"), filter_stopwords=True)


class DanishLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("da"), return_lemma=True)


class DanishLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("da"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class DanishLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("da"), return_lemma=True, filter_stopwords=True)


class PolishTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pl"))


class PolishFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("pl"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class PolishRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pl"), filter_stopwords=True)


class PolishLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pl"), return_lemma=True)


class PolishLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("pl"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class PolishLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("pl"), return_lemma=True, filter_stopwords=True)


class RomanianTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("ro"))


class RomanianFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("ro"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class RomanianRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("ro"), filter_stopwords=True)


class RomanianLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("ro"), return_lemma=True)


class RomanianLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("ro"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class RomanianLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("ro"), return_lemma=True, filter_stopwords=True)


class JapaneseTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("jp"))


class JapaneseFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("jp"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class JapaneseRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("jp"), filter_stopwords=True)


class JapaneseLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("jp"), return_lemma=True)


class JapaneseLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("jp"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class JapaneseLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("jp"), return_lemma=True, filter_stopwords=True)


class ChineseTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("zh"))


class ChineseFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("zh"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class ChineseRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("zh"), filter_stopwords=True)


class ChineseLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("zh"), return_lemma=True)


class ChineseLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("zh"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class ChineseLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("zh"), return_lemma=True, filter_stopwords=True)


class MultiTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("xx"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class MultiFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text, load_nlp_pipeline("xx"), filter_numbers=True, filter_punctuation=True, filter_short_tokens=True
        )


class MultiRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("xx"), filter_stopwords=True)


class MultiLemmatizeTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("xx"), return_lemma=True)


class MultiLemmatizeFilterTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(
            text,
            load_nlp_pipeline("xx"),
            return_lemma=True,
            filter_numbers=True,
            filter_punctuation=True,
            filter_short_tokens=True,
        )


class MultiLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
    def __call__(self, text):
        return process_text(text, load_nlp_pipeline("xx"), return_lemma=True, filter_stopwords=True)


class HFTokenizer(BaseTokenizer):
    def __init__(self, pretrained_model_name_or_path, **kwargs):
        super().__init__()
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        self.tokenizer = load_pretrained_hf_tokenizer(self.pretrained_model_name_or_path)

    def __call__(self, text):
        return self.tokenizer.encode(text, truncation=True)

    def get_vocab(self):
        return self.tokenizer.get_vocab()

    def get_pad_token(self) -> str:
        # HACK(geoffrey): gpt2 has no pad token. Recommendation is to use eos token instead.
        # https://github.com/huggingface/transformers/issues/2630#issuecomment-1290809338
        # https://github.com/huggingface/transformers/issues/2648#issuecomment-616177044
        if self.pretrained_model_name_or_path == "gpt2":
            self.tokenizer.pad_token = self.tokenizer.eos_token
        return self.tokenizer.pad_token

    def get_unk_token(self) -> str:
        return self.tokenizer.unk_token


tokenizer_registry = {
    # Torchscript-compatible tokenizers. Torchtext tokenizers are also available below (requires torchtext>=0.12.0).
    "space": SpaceStringToListTokenizer,
    "space_punct": SpacePunctuationStringToListTokenizer,
    "ngram": NgramTokenizer,
    # Tokenizers not compatible with torchscript
    "characters": CharactersToListTokenizer,
    "underscore": UnderscoreStringToListTokenizer,
    "comma": CommaStringToListTokenizer,
    "untokenized": UntokenizedStringToListTokenizer,
    "stripped": StrippedStringToListTokenizer,
    "english_tokenize": EnglishTokenizer,
    "english_tokenize_filter": EnglishFilterTokenizer,
    "english_tokenize_remove_stopwords": EnglishRemoveStopwordsTokenizer,
    "english_lemmatize": EnglishLemmatizeTokenizer,
    "english_lemmatize_filter": EnglishLemmatizeFilterTokenizer,
    "english_lemmatize_remove_stopwords": EnglishLemmatizeRemoveStopwordsTokenizer,
    "italian_tokenize": ItalianTokenizer,
    "italian_tokenize_filter": ItalianFilterTokenizer,
    "italian_tokenize_remove_stopwords": ItalianRemoveStopwordsTokenizer,
    "italian_lemmatize": ItalianLemmatizeTokenizer,
    "italian_lemmatize_filter": ItalianLemmatizeFilterTokenizer,
    "italian_lemmatize_remove_stopwords": ItalianLemmatizeRemoveStopwordsTokenizer,
    "spanish_tokenize": SpanishTokenizer,
    "spanish_tokenize_filter": SpanishFilterTokenizer,
    "spanish_tokenize_remove_stopwords": SpanishRemoveStopwordsTokenizer,
    "spanish_lemmatize": SpanishLemmatizeTokenizer,
    "spanish_lemmatize_filter": SpanishLemmatizeFilterTokenizer,
    "spanish_lemmatize_remove_stopwords": SpanishLemmatizeRemoveStopwordsTokenizer,
    "german_tokenize": GermanTokenizer,
    "german_tokenize_filter": GermanFilterTokenizer,
    "german_tokenize_remove_stopwords": GermanRemoveStopwordsTokenizer,
    "german_lemmatize": GermanLemmatizeTokenizer,
    "german_lemmatize_filter": GermanLemmatizeFilterTokenizer,
    "german_lemmatize_remove_stopwords": GermanLemmatizeRemoveStopwordsTokenizer,
    "french_tokenize": FrenchTokenizer,
    "french_tokenize_filter": FrenchFilterTokenizer,
    "french_tokenize_remove_stopwords": FrenchRemoveStopwordsTokenizer,
    "french_lemmatize": FrenchLemmatizeTokenizer,
    "french_lemmatize_filter": FrenchLemmatizeFilterTokenizer,
    "french_lemmatize_remove_stopwords": FrenchLemmatizeRemoveStopwordsTokenizer,
    "portuguese_tokenize": PortugueseTokenizer,
    "portuguese_tokenize_filter": PortugueseFilterTokenizer,
    "portuguese_tokenize_remove_stopwords": PortugueseRemoveStopwordsTokenizer,
    "portuguese_lemmatize": PortugueseLemmatizeTokenizer,
    "portuguese_lemmatize_filter": PortugueseLemmatizeFilterTokenizer,
    "portuguese_lemmatize_remove_stopwords": PortugueseLemmatizeRemoveStopwordsTokenizer,
    "dutch_tokenize": DutchTokenizer,
    "dutch_tokenize_filter": DutchFilterTokenizer,
    "dutch_tokenize_remove_stopwords": DutchRemoveStopwordsTokenizer,
    "dutch_lemmatize": DutchLemmatizeTokenizer,
    "dutch_lemmatize_filter": DutchLemmatizeFilterTokenizer,
    "dutch_lemmatize_remove_stopwords": DutchLemmatizeRemoveStopwordsTokenizer,
    "greek_tokenize": GreekTokenizer,
    "greek_tokenize_filter": GreekFilterTokenizer,
    "greek_tokenize_remove_stopwords": GreekRemoveStopwordsTokenizer,
    "greek_lemmatize": GreekLemmatizeTokenizer,
    "greek_lemmatize_filter": GreekLemmatizeFilterTokenizer,
    "greek_lemmatize_remove_stopwords": GreekLemmatizeRemoveStopwordsFilterTokenizer,
    "norwegian_tokenize": NorwegianTokenizer,
    "norwegian_tokenize_filter": NorwegianFilterTokenizer,
    "norwegian_tokenize_remove_stopwords": NorwegianRemoveStopwordsTokenizer,
    "norwegian_lemmatize": NorwegianLemmatizeTokenizer,
    "norwegian_lemmatize_filter": NorwegianLemmatizeFilterTokenizer,
    "norwegian_lemmatize_remove_stopwords": NorwegianLemmatizeRemoveStopwordsFilterTokenizer,
    "lithuanian_tokenize": LithuanianTokenizer,
    "lithuanian_tokenize_filter": LithuanianFilterTokenizer,
    "lithuanian_tokenize_remove_stopwords": LithuanianRemoveStopwordsTokenizer,
    "lithuanian_lemmatize": LithuanianLemmatizeTokenizer,
    "lithuanian_lemmatize_filter": LithuanianLemmatizeFilterTokenizer,
    "lithuanian_lemmatize_remove_stopwords": LithuanianLemmatizeRemoveStopwordsFilterTokenizer,
    "danish_tokenize": DanishTokenizer,
    "danish_tokenize_filter": DanishFilterTokenizer,
    "danish_tokenize_remove_stopwords": DanishRemoveStopwordsTokenizer,
    "danish_lemmatize": DanishLemmatizeTokenizer,
    "danish_lemmatize_filter": DanishLemmatizeFilterTokenizer,
    "danish_lemmatize_remove_stopwords": DanishLemmatizeRemoveStopwordsFilterTokenizer,
    "polish_tokenize": PolishTokenizer,
    "polish_tokenize_filter": PolishFilterTokenizer,
    "polish_tokenize_remove_stopwords": PolishRemoveStopwordsTokenizer,
    "polish_lemmatize": PolishLemmatizeTokenizer,
    "polish_lemmatize_filter": PolishLemmatizeFilterTokenizer,
    "polish_lemmatize_remove_stopwords": PolishLemmatizeRemoveStopwordsFilterTokenizer,
    "romanian_tokenize": RomanianTokenizer,
    "romanian_tokenize_filter": RomanianFilterTokenizer,
    "romanian_tokenize_remove_stopwords": RomanianRemoveStopwordsTokenizer,
    "romanian_lemmatize": RomanianLemmatizeTokenizer,
    "romanian_lemmatize_filter": RomanianLemmatizeFilterTokenizer,
    "romanian_lemmatize_remove_stopwords": RomanianLemmatizeRemoveStopwordsFilterTokenizer,
    "japanese_tokenize": JapaneseTokenizer,
    "japanese_tokenize_filter": JapaneseFilterTokenizer,
    "japanese_tokenize_remove_stopwords": JapaneseRemoveStopwordsTokenizer,
    "japanese_lemmatize": JapaneseLemmatizeTokenizer,
    "japanese_lemmatize_filter": JapaneseLemmatizeFilterTokenizer,
    "japanese_lemmatize_remove_stopwords": JapaneseLemmatizeRemoveStopwordsFilterTokenizer,
    "chinese_tokenize": ChineseTokenizer,
    "chinese_tokenize_filter": ChineseFilterTokenizer,
    "chinese_tokenize_remove_stopwords": ChineseRemoveStopwordsTokenizer,
    "chinese_lemmatize": ChineseLemmatizeTokenizer,
    "chinese_lemmatize_filter": ChineseLemmatizeFilterTokenizer,
    "chinese_lemmatize_remove_stopwords": ChineseLemmatizeRemoveStopwordsFilterTokenizer,
    "multi_tokenize": MultiTokenizer,
    "multi_tokenize_filter": MultiFilterTokenizer,
    "multi_tokenize_remove_stopwords": MultiRemoveStopwordsTokenizer,
    "multi_lemmatize": MultiLemmatizeTokenizer,
    "multi_lemmatize_filter": MultiLemmatizeFilterTokenizer,
    "multi_lemmatize_remove_stopwords": MultiLemmatizeRemoveStopwordsTokenizer,
}

try:
    import torchtext

    if torch.torch_version.TorchVersion(torchtext.__version__) >= (0, 12, 0):
        """torchtext 0.12.0 tokenizers.

        Only available with torchtext>=0.12.0.
        """

        class SentencePieceTokenizer(torch.nn.Module):
            def __init__(self, pretrained_model_name_or_path: Optional[str] = None, **kwargs):
                super().__init__()
                if pretrained_model_name_or_path is None:
                    pretrained_model_name_or_path = (
                        "https://download.pytorch.org/models/text/xlmr.sentencepiece.bpe.model"
                    )
                self.tokenizer = torchtext.transforms.SentencePieceTokenizer(
                    sp_model_path=pretrained_model_name_or_path
                )

            def forward(self, v: Union[str, List[str], torch.Tensor]):
                if isinstance(v, torch.Tensor):
                    raise ValueError(f"Unsupported input: {v}")
                return self.tokenizer(v)

        class _BPETokenizer(torch.nn.Module):
            """Superclass for tokenizers that use BPE, such as CLIPTokenizer and GPT2BPETokenizer."""

            def __init__(self, pretrained_model_name_or_path: str, vocab_file: str):
                super().__init__()
                self.str2idx, self.idx2str = self._init_vocab(vocab_file)
                self.tokenizer = self._init_tokenizer(pretrained_model_name_or_path, vocab_file)

            def _init_vocab(self, vocab_file: str) -> Dict[str, str]:
                """Loads the vocab from the vocab file."""
                str2idx = load_json(torchtext.utils.get_asset_local_path(vocab_file))
                _, idx2str = zip(*sorted((v, k) for k, v in str2idx.items()))
                return str2idx, idx2str

            def _init_tokenizer(self, pretrained_model_name_or_path: str, vocab_file: str) -> Any:
                """Initializes and returns the tokenizer."""
                raise NotImplementedError

            def forward(self, v: Union[str, List[str], torch.Tensor]) -> Any:
                """Implements forward pass for tokenizer.

                BPE tokenizers from torchtext return ids directly, which is inconsistent with the Ludwig tokenizer API.
                The below implementation works around this by converting the ids back to their original string tokens.
                """
                if isinstance(v, torch.Tensor):
                    raise ValueError(f"Unsupported input: {v}")

                inputs: List[str] = []
                # Ludwig calls map on List[str] objects, so we need to handle individual strings as well.
                if isinstance(v, str):
                    inputs.append(v)
                else:
                    inputs.extend(v)

                token_ids = self.tokenizer(inputs)
                assert torch.jit.isinstance(token_ids, List[List[str]])

                tokens = [[self.idx2str[int(unit_idx)] for unit_idx in sequence] for sequence in token_ids]
                return tokens[0] if isinstance(v, str) else tokens

            def get_vocab(self) -> Dict[str, str]:
                return self.str2idx

        class CLIPTokenizer(_BPETokenizer):
            def __init__(
                self, pretrained_model_name_or_path: Optional[str] = None, vocab_file: Optional[str] = None, **kwargs
            ):
                if pretrained_model_name_or_path is None:
                    pretrained_model_name_or_path = "http://download.pytorch.org/models/text/clip_merges.bpe"
                if vocab_file is None:
                    vocab_file = "http://download.pytorch.org/models/text/clip_encoder.json"
                super().__init__(pretrained_model_name_or_path, vocab_file)

            def _init_tokenizer(self, pretrained_model_name_or_path: str, vocab_file: str):
                return torchtext.transforms.CLIPTokenizer(
                    encoder_json_path=vocab_file, merges_path=pretrained_model_name_or_path
                )

        class GPT2BPETokenizer(_BPETokenizer):
            def __init__(
                self, pretrained_model_name_or_path: Optional[str] = None, vocab_file: Optional[str] = None, **kwargs
            ):
                if pretrained_model_name_or_path is None:
                    pretrained_model_name_or_path = "https://download.pytorch.org/models/text/gpt2_bpe_vocab.bpe"
                if vocab_file is None:
                    vocab_file = "https://download.pytorch.org/models/text/gpt2_bpe_encoder.json"
                super().__init__(pretrained_model_name_or_path, vocab_file)

            def _init_tokenizer(self, pretrained_model_name_or_path: str, vocab_file: str):
                return torchtext.transforms.GPT2BPETokenizer(
                    encoder_json_path=vocab_file, vocab_bpe_path=pretrained_model_name_or_path
                )

        tokenizer_registry.update(
            {
                "sentencepiece": SentencePieceTokenizer,
                "clip": CLIPTokenizer,
                "gpt2bpe": GPT2BPETokenizer,
            }
        )
        TORCHSCRIPT_COMPATIBLE_TOKENIZERS.update(TORCHTEXT_0_12_0_TOKENIZERS)
    else:
        raise ImportError(f"torchtext>=0.12.0 is required to use these tokenizers: {TORCHTEXT_0_12_0_TOKENIZERS}.")

except ImportError:
    pass


try:
    import torchtext

    torchtext_version = torch.torch_version.TorchVersion(torchtext.__version__)

    if torchtext_version >= (0, 13, 0):
        pass
    else:
        raise ImportError(f"torchtext>=0.13.0 is required to use these tokenizers: {TORCHTEXT_0_13_0_TOKENIZERS}.")

    class BERTTokenizer(torch.nn.Module):
        def __init__(
            self,
            vocab_file: Optional[str] = None,
            is_hf_tokenizer: Optional[bool] = False,
            hf_tokenizer_attrs: Optional[Dict[str, Any]] = None,
            **kwargs,
        ):
            super().__init__()

            if vocab_file is None:
                # If vocab_file not passed in, use default "bert-base-uncased" vocab and kwargs.
                kwargs = _get_bert_config("bert-base-uncased")
                vocab_file = kwargs["vocab_file"]
                vocab = self._init_vocab(vocab_file)
                hf_tokenizer_attrs = {
                    "pad_token": "[PAD]",
                    "unk_token": "[UNK]",
                    "sep_token_id": vocab["[SEP]"],
                    "cls_token_id": vocab["[CLS]"],
                }
            else:
                vocab = self._init_vocab(vocab_file)

            self.vocab = vocab

            self.is_hf_tokenizer = is_hf_tokenizer
            if self.is_hf_tokenizer:
                # Values used by Ludwig extracted from the corresponding HF model.
                self.pad_token = hf_tokenizer_attrs["pad_token"]  # Used as padding symbol
                self.unk_token = hf_tokenizer_attrs["unk_token"]  # Used as unknown symbol
                self.cls_token_id = hf_tokenizer_attrs["cls_token_id"]  # Used as start symbol. Only used if HF.
                self.sep_token_id = hf_tokenizer_attrs["sep_token_id"]  # Used as stop symbol. Only used if HF.
                self.never_split = hf_tokenizer_attrs["all_special_tokens"]
            else:
                self.pad_token = PADDING_SYMBOL
                self.unk_token = UNKNOWN_SYMBOL
                self.cls_token_id = None
                self.sep_token_id = None
                self.never_split = [UNKNOWN_SYMBOL]

            tokenizer_kwargs = {}
            if "do_lower_case" in kwargs:
                tokenizer_kwargs["do_lower_case"] = kwargs["do_lower_case"]
            if "strip_accents" in kwargs:
                tokenizer_kwargs["strip_accents"] = kwargs["strip_accents"]

            # Return tokens as raw tokens only if not being used as a HF tokenizer.
            self.return_tokens = not self.is_hf_tokenizer

            tokenizer_init_kwargs = {
                **tokenizer_kwargs,
                "vocab_path": vocab_file,
                "return_tokens": self.return_tokens,
            }
            if torchtext_version >= (0, 14, 0):
                # never_split kwarg added in torchtext 0.14.0
                tokenizer_init_kwargs["never_split"] = self.never_split

            self.tokenizer = torchtext.transforms.BERTTokenizer(**tokenizer_init_kwargs)

        def _init_vocab(self, vocab_file: str) -> Dict[str, int]:
            from transformers.models.bert.tokenization_bert import load_vocab

            return load_vocab(vocab_file)

        def forward(self, v: Union[str, List[str], torch.Tensor]) -> Any:
            """Implements forward pass for tokenizer.

            If the is_hf_tokenizer flag is set to True, then the output follows the HF convention, i.e. the output is an
            List[List[int]] of tokens and the cls and sep tokens are automatically added as the start and stop symbols.

            If the is_hf_tokenizer flag is set to False, then the output follows the Ludwig convention, i.e. the output
            is a List[List[str]] of tokens.
            """
            if isinstance(v, torch.Tensor):
                raise ValueError(f"Unsupported input: {v}")

            inputs: List[str] = []
            # Ludwig calls map on List[str] objects, so we need to handle individual strings as well.
            if isinstance(v, str):
                inputs.append(v)
            else:
                inputs.extend(v)

            if self.is_hf_tokenizer:
                token_ids_str = self.tokenizer(inputs)
                assert torch.jit.isinstance(token_ids_str, List[List[str]])
                # Must cast token_ids to ints because they are used directly as indices.
                token_ids: List[List[int]] = []
                for token_ids_str_i in token_ids_str:
                    token_ids_i = [int(token_id_str) for token_id_str in token_ids_str_i]
                    token_ids_i = self._add_special_token_ids(token_ids_i)
                    token_ids.append(token_ids_i)
                return token_ids[0] if isinstance(v, str) else token_ids

            tokens = self.tokenizer(inputs)
            assert torch.jit.isinstance(tokens, List[List[str]])
            return tokens[0] if isinstance(v, str) else tokens

        def get_vocab(self) -> Dict[str, int]:
            return self.vocab

        def get_pad_token(self) -> str:
            return self.pad_token

        def get_unk_token(self) -> str:
            return self.unk_token

        def _add_special_token_ids(self, token_ids: List[int]) -> List[int]:
            """Adds special token ids to the token_ids list."""
            if torch.jit.isinstance(self.cls_token_id, int) and torch.jit.isinstance(self.sep_token_id, int):
                token_ids.insert(0, self.cls_token_id)
                token_ids.append(self.sep_token_id)
            return token_ids

    tokenizer_registry.update(
        {
            "bert": BERTTokenizer,
        }
    )
    TORCHSCRIPT_COMPATIBLE_TOKENIZERS.update(TORCHTEXT_0_13_0_TOKENIZERS)

except ImportError:
    pass


def get_hf_tokenizer(pretrained_model_name_or_path, **kwargs):
    """Gets a potentially torchscript-compatible tokenizer that follows HF convention.

    Args:
        pretrained_model_name_or_path: Name of the model in the HF repo. Example: "bert-base-uncased".
    Returns:
        A torchscript-able HF tokenizer if it is available. Else, returns vanilla HF tokenizer.
    """
    from transformers import BertTokenizer

    hf_name = pretrained_model_name_or_path
    # use_fast=False to leverage python class inheritance
    # cannot tokenize HF tokenizers directly because HF lacks strict typing and List[str] cannot be traced
    hf_tokenizer = load_pretrained_hf_tokenizer(hf_name, use_fast=False)

    torchtext_tokenizer = None
    if "bert" in TORCHSCRIPT_COMPATIBLE_TOKENIZERS and isinstance(hf_tokenizer, BertTokenizer):
        tokenizer_kwargs = _get_bert_config(hf_name)
        torchtext_tokenizer = BERTTokenizer(
            **tokenizer_kwargs,
            is_hf_tokenizer=True,
            hf_tokenizer_attrs={
                "pad_token": hf_tokenizer.pad_token,
                "unk_token": hf_tokenizer.unk_token,
                "cls_token_id": hf_tokenizer.cls_token_id,
                "sep_token_id": hf_tokenizer.sep_token_id,
                "all_special_tokens": hf_tokenizer.all_special_tokens,
            },
        )

    use_torchtext = torchtext_tokenizer is not None
    if use_torchtext:
        # If a torchtext tokenizer is instantiable, tenatively we will use it. However,
        # if the tokenizer does not pass (lightweight) validation, then we will fall back to the vanilla HF tokenizer.
        # TODO(geoffrey): can we better validate tokenizer parity before swapping in the TorchText tokenizer?
        # Samples from https://github.com/huggingface/transformers/blob/main/tests/models/bert/test_tokenization_bert.py
        for sample_input in HF_TOKENIZER_SAMPLE_INPUTS:
            hf_output = hf_tokenizer.encode(sample_input)
            tt_output = torchtext_tokenizer(sample_input)
            if hf_output != tt_output:
                use_torchtext = False
                logger.warning("Falling back to HuggingFace tokenizer because TorchText tokenizer failed validation.")
                logger.warning(f"Sample input: {sample_input}\nHF output: {hf_output}\nTT output: {tt_output}")
                break

    if use_torchtext:
        logger.info(f"Loaded TorchText implementation of {hf_name} tokenizer")
        return torchtext_tokenizer
    else:
        # If hf_name does not have a torchtext equivalent implementation, load the
        # HuggingFace implementation.
        logger.info(f"Loaded HuggingFace implementation of {hf_name} tokenizer")
        return HFTokenizer(hf_name)


def _get_bert_config(hf_name):
    """Gets configs from BERT tokenizers in HuggingFace.

    `vocab_file` is required for BERT tokenizers. `tokenizer_config.json` are optional keyword arguments used to
    initialize the tokenizer object. If no `tokenizer_config.json` is found, then we instantiate the tokenizer with
    default arguments.
    """
    from transformers.utils.hub import cached_path, EntryNotFoundError

    vocab_file = cached_path(f"https://huggingface.co/{hf_name}/resolve/main/vocab.txt")

    try:
        tokenizer_config = load_json(
            cached_path(f"https://huggingface.co/{hf_name}/resolve/main/tokenizer_config.json")
        )
    except EntryNotFoundError:
        tokenizer_config = {}

    return {"vocab_file": vocab_file, **tokenizer_config}


tokenizer_registry.update(
    {
        "hf_tokenizer": get_hf_tokenizer,
    }
)


def get_tokenizer_from_registry(tokenizer_name: str) -> torch.nn.Module:
    """Returns the appropriate tokenizer from the tokenizer registry.

    Raises a KeyError if a tokenizer that does not exist in the registry is requested, with additional help text if the
    requested tokenizer would be available for a different version of torchtext.
    """
    if tokenizer_name in tokenizer_registry:
        return tokenizer_registry[tokenizer_name]

    if (
        torch.torch_version.TorchVersion(torchtext.__version__) < (0, 12, 0)
        and tokenizer_name in TORCHTEXT_0_12_0_TOKENIZERS
    ):
        raise KeyError(
            f"torchtext>=0.12.0 is not installed, so '{tokenizer_name}' and the following tokenizers are not "
            f"available: {TORCHTEXT_0_12_0_TOKENIZERS}"
        )

    if (
        torch.torch_version.TorchVersion(torchtext.__version__) < (0, 13, 0)
        and tokenizer_name in TORCHTEXT_0_13_0_TOKENIZERS
    ):
        raise KeyError(
            f"torchtext>=0.13.0 is not installed, so '{tokenizer_name}' and the following tokenizers are not "
            f"available: {TORCHTEXT_0_13_0_TOKENIZERS}"
        )

    # Tokenizer does not exist or is unavailable.
    raise KeyError(f"Invalid tokenizer name: '{tokenizer_name}'. Available tokenizers: {tokenizer_registry.keys()}")



import ludwig.datasets


def test_default_model_config(tmpdir):
    titanic_configs = ludwig.datasets.model_configs_for_dataset("titanic")
    assert len(titanic_configs) > 0

    titanic = ludwig.datasets.get_dataset("titanic", cache_dir=tmpdir)
    assert titanic.default_model_config is not None

    assert titanic.default_model_config == titanic_configs["default"]


def test_best_model_config(tmpdir):
    higgs_configs = ludwig.datasets.model_configs_for_dataset("higgs")
    assert len(higgs_configs) > 0

    higgs = ludwig.datasets.get_dataset("higgs", cache_dir=tmpdir)
    assert higgs.default_model_config is not None
    assert higgs.best_model_config is not None

    assert higgs.default_model_config == higgs_configs["default"]
    assert higgs.best_model_config == higgs_configs["best"]


def test_dataset_has_no_model_configs(tmpdir):
    bbc_news_configs = ludwig.datasets.model_configs_for_dataset("bbcnews")
    assert len(bbc_news_configs) == 0

    bbcnews = ludwig.datasets.get_dataset("bbcnews", cache_dir=tmpdir)
    assert bbcnews.default_model_config is None



import os
from typing import Type

import pytest
from transformers import BertModel, BertTokenizer

from ludwig.encoders.text_encoders import BERTEncoder
from ludwig.utils.hf_utils import load_pretrained_hf_model


@pytest.mark.parametrize(
    ("model", "name"),
    [
        (BertModel, BERTEncoder.DEFAULT_MODEL_NAME),
        (BertTokenizer, "bert-base-uncased"),
    ],
)
def test_load_hf_model(model: Type, name: str, tmpdir: os.PathLike):
    """Ensure that the HF models used in ludwig download correctly."""
    cache_dir = os.path.join(tmpdir, name.replace(os.path.sep, "_") if name else str(model.__name__))
    os.makedirs(cache_dir, exist_ok=True)
    loaded_model = load_pretrained_hf_model(model, name, cache_dir=cache_dir, force_download=True)
    assert isinstance(loaded_model, model)
    assert os.listdir(cache_dir)



"""Bisection implementation of alpha-entmax (Peters et al., 2019).

Backward pass wrt alpha per (Correia et al., 2019). See https://arxiv.org/pdf/1905.05702 for detailed description.
"""
# Author: Goncalo M Correia
# Author: Ben Peters
# Author: Vlad Niculae <vlad@vene.ro>

import torch
import torch.nn as nn
from torch.autograd import Function


class EntmaxBisectFunction(Function):
    @classmethod
    def _gp(cls, x, alpha):
        return x ** (alpha - 1)

    @classmethod
    def _gp_inv(cls, y, alpha):
        return y ** (1 / (alpha - 1))

    @classmethod
    def _p(cls, X, alpha):
        return cls._gp_inv(torch.clamp(X, min=0), alpha)

    @classmethod
    def forward(cls, ctx, X, alpha=1.5, dim=-1, n_iter=50, ensure_sum_one=True):
        p_m, backward_kwargs = _entmax_bisect_forward(X, alpha, dim, n_iter, ensure_sum_one, cls)

        ctx.alpha = backward_kwargs["alpha"]
        ctx.dim = backward_kwargs["dim"]
        ctx.save_for_backward(p_m)
        return p_m

    @classmethod
    def backward(cls, ctx, dY):
        (Y,) = ctx.saved_tensors

        gppr = torch.where(Y > 0, Y ** (2 - ctx.alpha), Y.new_zeros(1))

        dX = dY * gppr
        q = dX.sum(ctx.dim) / gppr.sum(ctx.dim)
        q = q.unsqueeze(ctx.dim)
        dX -= q * gppr

        d_alpha = None
        if ctx.needs_input_grad[1]:

            # alpha gradient computation
            # d_alpha = (partial_y / partial_alpha) * dY
            # NOTE: ensure alpha is not close to 1
            # since there is an indetermination
            # batch_size, _ = dY.shape

            # shannon terms
            S = torch.where(Y > 0, Y * torch.log(Y), Y.new_zeros(1))
            # shannon entropy
            ent = S.sum(ctx.dim).unsqueeze(ctx.dim)
            Y_skewed = gppr / gppr.sum(ctx.dim).unsqueeze(ctx.dim)

            d_alpha = dY * (Y - Y_skewed) / ((ctx.alpha - 1) ** 2)
            d_alpha -= dY * (S - Y_skewed * ent) / (ctx.alpha - 1)
            d_alpha = d_alpha.sum(ctx.dim).unsqueeze(ctx.dim)

        return dX, d_alpha, None, None, None


def _entmax_bisect_forward(X, alpha, dim, n_iter, ensure_sum_one, cls=EntmaxBisectFunction):
    if not isinstance(alpha, torch.Tensor):
        alpha = torch.tensor(alpha, dtype=X.dtype, device=X.device)

    alpha_shape = list(X.shape)
    alpha_shape[dim] = 1
    alpha = alpha.expand(*alpha_shape)

    d = X.shape[dim]

    max_val, _ = X.max(dim=dim, keepdim=True)
    X = X * (alpha - 1)
    max_val = max_val * (alpha - 1)

    # Note: when alpha < 1, tau_lo > tau_hi. This still works since dm < 0.
    tau_lo = max_val - cls._gp(1, alpha)
    tau_hi = max_val - cls._gp(1 / d, alpha)

    f_lo = cls._p(X - tau_lo, alpha).sum(dim) - 1

    dm = tau_hi - tau_lo

    for it in range(n_iter):

        dm /= 2
        tau_m = tau_lo + dm
        p_m = cls._p(X - tau_m, alpha)
        f_m = p_m.sum(dim) - 1

        mask = (f_m * f_lo >= 0).unsqueeze(dim)
        tau_lo = torch.where(mask, tau_m, tau_lo)

    if ensure_sum_one:
        p_m /= p_m.sum(dim=dim).unsqueeze(dim=dim)

    return p_m, {"alpha": alpha, "dim": dim}


# slightly more efficient special case for sparsemax
class SparsemaxBisectFunction(EntmaxBisectFunction):
    @classmethod
    def _gp(cls, x, alpha):
        return x

    @classmethod
    def _gp_inv(cls, y, alpha):
        return y

    @classmethod
    def _p(cls, x, alpha):
        return torch.clamp(x, min=0)

    @classmethod
    def forward(cls, ctx, X, dim=-1, n_iter=50, ensure_sum_one=True):
        p_m, backward_kwargs = _sparsemax_bisect_forward(X, dim, n_iter, ensure_sum_one)

        ctx.alpha = backward_kwargs["alpha"]
        ctx.dim = backward_kwargs["dim"]
        ctx.save_for_backward(p_m)
        return p_m

    @classmethod
    def backward(cls, ctx, dY):
        (Y,) = ctx.saved_tensors
        gppr = (Y > 0).to(dtype=dY.dtype)
        dX = dY * gppr
        q = dX.sum(ctx.dim) / gppr.sum(ctx.dim)
        q = q.unsqueeze(ctx.dim)
        dX -= q * gppr
        return dX, None, None, None


def _sparsemax_bisect_forward(X, dim, n_iter, ensure_sum_one):
    return _entmax_bisect_forward(X, alpha=2, dim=dim, n_iter=50, ensure_sum_one=True, cls=SparsemaxBisectFunction)


def entmax_bisect(X, alpha=1.5, dim=-1, n_iter=50, ensure_sum_one=True, training=True):
    """alpha-entmax: normalizing sparse transform (a la softmax).

    Solves the optimization problem:

        max_p <x, p> - H_a(p)    s.t.    p >= 0, sum(p) == 1.

    where H_a(p) is the Tsallis alpha-entropy with custom alpha >= 1,
    using a bisection (root finding, binary search) algorithm.

    This function is differentiable with respect to both X and alpha.

    Parameters
    ----------
    X : torch.Tensor
        The input tensor.

    alpha : float or torch.Tensor
        Tensor of alpha parameters (> 1) to use. If scalar
        or python float, the same value is used for all rows, otherwise,
        it must have shape (or be expandable to)
        alpha.shape[j] == (X.shape[j] if j != dim else 1)
        A value of alpha=2 corresponds to sparsemax, and alpha=1 would in theory recover
        softmax. For numeric reasons, this algorithm does not work with `alpha=1`: if you
        want softmax, we recommend `torch.nn.softmax`.

    dim : int
        The dimension along which to apply alpha-entmax.

    n_iter : int
        Number of bisection iterations. For float32, 24 iterations should
        suffice for machine precision.

    ensure_sum_one : bool,
        Whether to divide the result by its sum. If false, the result might
        sum to close but not exactly 1, which might cause downstream problems.

    Returns
    -------
    P : torch tensor, same shape as X
        The projection result, such that P.sum(dim=dim) == 1 elementwise.
    """
    # Avoids call to custom autograd.Function during eval to ensure torchscript compatibility
    # custom autograd.Function is not scriptable: https://github.com/pytorch/pytorch/issues/22329#issuecomment-506608053
    if not training:
        output, _ = _entmax_bisect_forward(X, alpha, dim, n_iter, ensure_sum_one)
        return output
    return EntmaxBisectFunction.apply(X, alpha, dim, n_iter, ensure_sum_one)


def sparsemax_bisect(X, dim=-1, n_iter=50, ensure_sum_one=True, training=True):
    """sparsemax: normalizing sparse transform (a la softmax), via bisection.

    Solves the projection:

        min_p ||x - p||_2   s.t.    p >= 0, sum(p) == 1.

    Parameters
    ----------
    X : torch.Tensor
        The input tensor.

    dim : int
        The dimension along which to apply sparsemax.

    n_iter : int
        Number of bisection iterations. For float32, 24 iterations should
        suffice for machine precision.

    ensure_sum_one : bool,
        Whether to divide the result by its sum. If false, the result might
        sum to close but not exactly 1, which might cause downstream problems.

    Note: This function does not yet support normalizing along anything except
    the last dimension. Please use transposing and views to achieve more
    general behavior.

    Returns
    -------
    P : torch tensor, same shape as X
        The projection result, such that P.sum(dim=dim) == 1 elementwise.
    """
    # Avoids call to custom autograd.Function during eval to ensure torchscript compatibility
    # custom autograd.Function is not scriptable: https://github.com/pytorch/pytorch/issues/22329#issuecomment-506608053
    if not training:
        output, _ = _sparsemax_bisect_forward(X, dim, n_iter, ensure_sum_one)
        return output
    return SparsemaxBisectFunction.apply(X, dim, n_iter, ensure_sum_one)


class SparsemaxBisect(nn.Module):
    def __init__(self, dim=-1, n_iter=None):
        """sparsemax: normalizing sparse transform (a la softmax) via bisection

        Solves the projection:

            min_p ||x - p||_2   s.t.    p >= 0, sum(p) == 1.

        Parameters
        ----------
        dim : int
            The dimension along which to apply sparsemax.

        n_iter : int
            Number of bisection iterations. For float32, 24 iterations should
            suffice for machine precision.
        """
        self.dim = dim
        self.n_iter = n_iter
        super().__init__()

    def forward(self, X):
        return sparsemax_bisect(X, dim=self.dim, n_iter=self.n_iter, training=self.training)


class EntmaxBisect(nn.Module):
    def __init__(self, alpha=1.5, dim=-1, n_iter=50):
        """alpha-entmax: normalizing sparse map (a la softmax) via bisection.

        Solves the optimization problem:

            max_p <x, p> - H_a(p)    s.t.    p >= 0, sum(p) == 1.

        where H_a(p) is the Tsallis alpha-entropy with custom alpha >= 1,
        using a bisection (root finding, binary search) algorithm.

        Parameters
        ----------
        alpha : float or torch.Tensor
            Tensor of alpha parameters (> 1) to use. If scalar
            or python float, the same value is used for all rows, otherwise,
            it must have shape (or be expandable to)
            alpha.shape[j] == (X.shape[j] if j != dim else 1)
            A value of alpha=2 corresponds to sparsemax; and alpha=1 would in theory recover
            softmax. For numeric reasons, this algorithm does not work with `alpha=1`; if you
            want softmax, we recommend `torch.nn.softmax`.

        dim : int
            The dimension along which to apply alpha-entmax.

        n_iter : int
            Number of bisection iterations. For float32, 24 iterations should
            suffice for machine precision.

        """
        self.dim = dim
        self.n_iter = n_iter
        self.alpha = alpha
        super().__init__()

    def forward(self, X):
        return entmax_bisect(X, alpha=self.alpha, dim=self.dim, n_iter=self.n_iter, training=self.training)



import logging
import os
import queue
import threading

from ludwig.api_annotations import DeveloperAPI, PublicAPI
from ludwig.callbacks import Callback
from ludwig.constants import TRAINER
from ludwig.globals import MODEL_HYPERPARAMETERS_FILE_NAME, TRAIN_SET_METADATA_FILE_NAME
from ludwig.types import TrainingSetMetadataDict
from ludwig.utils.data_utils import chunk_dict, flatten_dict, save_json, to_json_dict
from ludwig.utils.package_utils import LazyLoader

mlflow = LazyLoader("mlflow", globals(), "mlflow")

logger = logging.getLogger(__name__)


def _get_runs(experiment_id: str):
    return mlflow.tracking.client.MlflowClient().search_runs([experiment_id])


@DeveloperAPI
def get_or_create_experiment_id(experiment_name, artifact_uri: str = None):
    """Gets experiment id from mlflow."""
    experiment = mlflow.get_experiment_by_name(experiment_name)
    if experiment is not None:
        return experiment.experiment_id
    return mlflow.create_experiment(name=experiment_name, artifact_location=artifact_uri)


# Included for backwards compatibility, Deprecated.
# TODO(daniel): delete this.
_get_or_create_experiment_id = get_or_create_experiment_id


@PublicAPI
class MlflowCallback(Callback):
    def __init__(self, tracking_uri=None, log_artifacts: bool = True):
        if tracking_uri:
            mlflow.set_tracking_uri(tracking_uri)
        self.tracking_uri = mlflow.get_tracking_uri()

        active_run = mlflow.active_run()
        if active_run is not None:
            # Use experiment already set in the current environment
            self.run = active_run
            self.experiment_id = self.run.info.experiment_id
            self.experiment_name = mlflow.get_experiment(self.experiment_id).name
            self.external_run = True
        else:
            # Will create an experiment at training time
            self.run = None
            self.experiment_id = None
            self.experiment_name = None
            self.external_run = False

        self.run_ended = False
        self.training_set_metadata = None
        self.config = None
        self.save_in_background = True
        self.save_fn = None
        self.save_thread = None
        self.log_artifacts = log_artifacts

    def get_experiment_id(self, experiment_name):
        return get_or_create_experiment_id(experiment_name)

    def on_preprocess_end(
        self,
        training_set: "Dataset",  # noqa
        validation_set: "Dataset",  # noqa
        test_set: "Dataset",  # noqa
        training_set_metadata: TrainingSetMetadataDict,
    ):
        self.training_set_metadata = training_set_metadata

    def on_hyperopt_init(self, experiment_name):
        self.experiment_id = self.get_experiment_id(experiment_name)
        self.experiment_name = experiment_name

    def on_hyperopt_trial_start(self, parameters):
        # Filter out mlflow params like tracking URI, experiment ID, etc.
        params = {k: v for k, v in parameters.items() if k != "mlflow"}
        self._log_params({"hparam": params})

        # TODO(travis): figure out a good way to support this. The problem with
        # saving artifacts in the background with hyperopt is early stopping. If
        # the scheduler decides to terminate a process, then currently there's no
        # mechanism to detect this a "flush" the queue of pending writes before
        # stopping. Should work with Ray Tune team to come up with a solution.
        self.save_in_background = False

    def on_train_init(self, base_config, experiment_name, output_directory, resume_directory, **kwargs):
        # Experiment may already have been set during hyperopt init, in
        # which case we don't want to create a new experiment / run, as
        # this should be handled by the executor.
        if self.experiment_id is None:
            mlflow.end_run()
            self.experiment_id = self.get_experiment_id(experiment_name)
            self.experiment_name = experiment_name

        active_run = mlflow.active_run()
        if active_run is not None:
            # Currently active run started by Ray Tune MLflow mixin or external run
            self.run = active_run
        else:
            run_id = None
            if resume_directory is not None:
                previous_runs = _get_runs(self.experiment_id)
                if len(previous_runs) > 0:
                    run_id = previous_runs[0].info.run_id
            if run_id is not None:
                self.run = mlflow.start_run(run_id=run_id)
            else:
                run_name = os.path.basename(output_directory)
                self.run = mlflow.start_run(experiment_id=self.experiment_id, run_name=run_name)

        self.log_config(base_config)

    def log_config(self, config):
        if self.log_artifacts:
            mlflow.log_dict(to_json_dict(config), "config.yaml")

    def on_train_start(self, config, **kwargs):
        self.config = config
        self._log_params({TRAINER: config[TRAINER]})

    def on_train_end(self, output_directory):
        if self.log_artifacts:
            _log_artifacts(output_directory)
        if self.run is not None and not self.external_run:
            # Only end runs managed internally to this callback
            mlflow.end_run()
            self.run_ended = True

    def on_trainer_train_setup(self, trainer, save_path, is_coordinator):
        if not is_coordinator:
            return

        # When running on a remote worker, the model metadata files will only have been
        # saved to the driver process, so re-save it here before uploading.
        training_set_metadata_path = os.path.join(save_path, TRAIN_SET_METADATA_FILE_NAME)
        if not os.path.exists(training_set_metadata_path):
            save_json(training_set_metadata_path, self.training_set_metadata)

        model_hyperparameters_path = os.path.join(save_path, MODEL_HYPERPARAMETERS_FILE_NAME)
        if not os.path.exists(model_hyperparameters_path):
            save_json(model_hyperparameters_path, self.config)

        if self.save_in_background:
            save_queue = queue.Queue()
            self.save_fn = lambda args: save_queue.put(args)
            self.save_thread = threading.Thread(target=_log_mlflow_loop, args=(save_queue, self.log_artifacts))
            self.save_thread.start()
        else:
            self.save_fn = lambda args: _log_mlflow(*args, self.log_artifacts)

    def on_eval_end(self, trainer, progress_tracker, save_path):
        self.save_fn((progress_tracker.log_metrics(), progress_tracker.steps, save_path, True))

    def on_trainer_train_teardown(self, trainer, progress_tracker, save_path, is_coordinator):
        if is_coordinator:
            self.save_fn((progress_tracker.log_metrics(), progress_tracker.steps, save_path, False))
            if self.save_thread is not None:
                self.save_thread.join()

    def on_visualize_figure(self, fig):
        # TODO: need to also include a filename for this figure
        # mlflow.log_figure(fig)
        pass

    def prepare_ray_tune(self, train_fn, tune_config, tune_callbacks):
        from ray.tune.integration.mlflow import mlflow_mixin

        return mlflow_mixin(train_fn), {
            **tune_config,
            "mlflow": {
                "experiment_id": self.experiment_id,
                "experiment_name": self.experiment_name,
                "tracking_uri": mlflow.get_tracking_uri(),
            },
        }

    def _log_params(self, params):
        flat_params = flatten_dict(params)
        for chunk in chunk_dict(flat_params, chunk_size=100):
            mlflow.log_params(chunk)

    def __setstate__(self, d):
        self.__dict__ = d
        if self.tracking_uri:
            mlflow.set_tracking_uri(self.tracking_uri)
        if self.run and not self.run_ended:
            # Run has already been set, but may not be active due to training workers running in a separate
            # process, so resume the run
            mlflow.end_run()
            self.run = mlflow.start_run(run_id=self.run.info.run_id, experiment_id=self.run.info.experiment_id)


def _log_mlflow_loop(q: queue.Queue, log_artifacts: bool = True):
    should_continue = True
    while should_continue:
        elem = q.get()
        log_metrics, steps, save_path, should_continue = elem
        mlflow.log_metrics(log_metrics, step=steps)

        if not q.empty():
            # in other words, don't bother saving the model artifacts
            # if we're about to do it again
            continue

        if log_artifacts:
            _log_model(save_path)


def _log_mlflow(log_metrics, steps, save_path, should_continue, log_artifacts: bool = True):
    mlflow.log_metrics(log_metrics, step=steps)
    if log_artifacts:
        _log_model(save_path)


def _log_artifacts(output_directory):
    for fname in os.listdir(output_directory):
        lpath = os.path.join(output_directory, fname)
        if fname == "model":
            _log_model(lpath)
        else:
            mlflow.log_artifact(lpath)


def _log_model(lpath):
    # Lazy import to avoid requiring this package
    from ludwig.contribs.mlflow.model import log_saved_model

    log_saved_model(lpath)



import logging
from collections import OrderedDict
from typing import Dict

from tabulate import tabulate

from ludwig.constants import COMBINED, LOSS
from ludwig.utils.metric_utils import TrainerMetric

logger = logging.getLogger(__name__)


class MetricsPrintedTable:
    """Maintains a table data structure used for logging metrics.

    ╒════════════╤════════════╤════════╤═════════════╤══════════╤═══════════╕
    │ Survived   │   accuracy │   loss │   precision │   recall │   roc_auc │
    ╞════════════╪════════════╪════════╪═════════════╪══════════╪═══════════╡
    │ train      │     0.7420 │ 0.7351 │      0.7107 │   0.5738 │    0.7659 │
    ├────────────┼────────────┼────────┼─────────────┼──────────┼───────────┤
    │ validation │     0.7079 │ 0.9998 │      0.6061 │   0.6061 │    0.7354 │
    ├────────────┼────────────┼────────┼─────────────┼──────────┼───────────┤
    │ test       │     0.7360 │ 0.7620 │      0.6667 │   0.5538 │    0.7358 │
    ╘════════════╧════════════╧════════╧═════════════╧══════════╧═══════════╛
    ╒════════════╤════════╕
    │ combined   │   loss │
    ╞════════════╪════════╡
    │ train      │ 0.7351 │
    ├────────────┼────────┤
    │ validation │ 0.9998 │
    ├────────────┼────────┤
    │ test       │ 0.7620 │
    ╘════════════╧════════╛
    """

    def __init__(self, output_features: Dict[str, "OutputFeature"]):  # noqa
        self.printed_table = OrderedDict()
        for output_feature_name, output_feature in output_features.items():
            self.printed_table[output_feature_name] = [[output_feature_name] + output_feature.metric_names]
        self.printed_table[COMBINED] = [[COMBINED, LOSS]]

        # Establish the printed table's order of metrics (used for appending metrics in the right order).
        self.metrics_headers = {}
        for output_feature_name in output_features.keys():
            # [0]: The header is the first row, which contains names of metrics.
            # [1:]: Skip the first column as it's just the name of the output feature, not an actual metric name.
            self.metrics_headers[output_feature_name] = self.printed_table[output_feature_name][0][1:]
        self.metrics_headers[COMBINED] = [LOSS]

    def add_metrics_to_printed_table(self, metrics_log: Dict[str, Dict[str, TrainerMetric]], split_name: str):
        """Add metrics to tables by the order of the table's metric header."""
        for output_feature_name, output_feature_metrics in metrics_log.items():
            printed_metrics = []
            for metric_name in self.metrics_headers[output_feature_name]:
                # Metrics may be missing if should_evaluate_train is False.
                if metric_name in output_feature_metrics and output_feature_metrics[metric_name]:
                    printed_metrics.append(output_feature_metrics[metric_name][-1][-1])
                else:
                    printed_metrics.append("")
            self.printed_table[output_feature_name].append([split_name] + printed_metrics)

    def log_info(self):
        for output_feature, table in self.printed_table.items():
            logger.info(tabulate(table, headers="firstrow", tablefmt="fancy_grid", floatfmt=".4f"))



from typing import Dict, Optional

import numpy as np

from ludwig.constants import SPLIT
from ludwig.types import FeatureConfigDict, PreprocessingConfigDict
from ludwig.utils.dataframe_utils import is_dask_series_or_df
from ludwig.utils.types import DataFrame


def convert_to_dict(
    predictions: DataFrame,
    output_features: Dict[str, FeatureConfigDict],
    backend: Optional["Backend"] = None,  # noqa: F821
):
    """Convert predictions from DataFrame format to a dictionary."""
    output = {}
    for of_name, output_feature in output_features.items():
        feature_keys = {k for k in predictions.columns if k.startswith(of_name)}
        feature_dict = {}
        for key in feature_keys:
            subgroup = key[len(of_name) + 1 :]

            values = predictions[key]
            if is_dask_series_or_df(values, backend):
                values = values.compute()
            try:
                values = np.stack(values.to_numpy())
            except ValueError:
                values = values.to_list()

            feature_dict[subgroup] = values
        output[of_name] = feature_dict
    return output


def set_fixed_split(preprocessing_params: PreprocessingConfigDict) -> PreprocessingConfigDict:
    """Sets the split policy explicitly to a fixed split.

    This potentially overrides the split configuration that the user set or what came from schema defaults.
    """

    return {
        **preprocessing_params,
        "split": {
            "type": "fixed",
            "column": SPLIT,
        },
    }



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Dict, List, Optional

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import H3
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.modules.embedding_modules import Embed, EmbedSequence
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.modules.initializer_modules import get_initializer
from ludwig.modules.recurrent_modules import RecurrentStack
from ludwig.modules.reduction_modules import SequenceReducer
from ludwig.schema.encoders.h3_encoders import H3EmbedConfig, H3RNNConfig, H3WeightedSumConfig
from ludwig.utils import torch_utils

logger = logging.getLogger(__name__)

# TODO: Share this with h3_feature.H3_VECTOR_LENGTH
H3_INPUT_SIZE = 19


@DeveloperAPI
@register_encoder("embed", H3)
class H3Embed(Encoder):
    def __init__(
        self,
        embedding_size: int = 10,
        embeddings_on_cpu: bool = False,
        fc_layers: Optional[List] = None,
        num_fc_layers: int = 0,
        output_size: int = 10,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: str = None,
        norm_params: Dict = None,
        activation: str = "relu",
        dropout: float = 0,
        reduce_output: str = "sum",
        encoder_config=None,
        **kwargs,
    ):
        """
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for
               `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memory and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.embedding_size = embedding_size
        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(reduce_mode=reduce_output)

        logger.debug("  mode Embed")
        self.embed_mode = Embed(
            [str(i) for i in range(3)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=True,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  edge Embed")
        self.embed_edge = Embed(
            [str(i) for i in range(7)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=True,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  resolution Embed")
        self.embed_resolution = Embed(
            [str(i) for i in range(16)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=True,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  base cell Embed")
        self.embed_base_cell = Embed(
            [str(i) for i in range(122)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=True,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  cells Embed")
        self.embed_cells = EmbedSequence(
            [str(i) for i in range(8)],
            embedding_size,
            max_sequence_length=(H3_INPUT_SIZE - 4),
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            force_embedding_size=True,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=embedding_size,
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        :param inputs: The input vector fed into the encoder.
               Shape: [batch x H3_INPUT_SIZE], type torch.int8
        :type inputs: Tensor
        """
        input_vector = inputs.int()

        # ================ Embeddings ================
        embedded_mode = self.embed_mode(input_vector[:, 0:1]).unsqueeze(1)
        embedded_edge = self.embed_edge(input_vector[:, 1:2]).unsqueeze(1)
        embedded_resolution = self.embed_resolution(input_vector[:, 2:3]).unsqueeze(1)
        embedded_base_cell = self.embed_base_cell(input_vector[:, 3:4]).unsqueeze(1)
        embedded_cells = self.embed_cells(input_vector[:, 4:])

        # ================ Masking ================
        # Mask out cells beyond the resolution of interest.
        resolution = input_vector[:, 2]
        mask = torch.unsqueeze(torch_utils.sequence_mask(resolution, 15), dim=-1).float()
        # Batch size X 15(max resolution) X embedding size
        masked_embedded_cells = embedded_cells * mask

        # ================ Reduce ================
        # Batch size X H3_INPUT_SIZE X embedding size
        concatenated = torch.cat(
            [embedded_mode, embedded_edge, embedded_resolution, embedded_base_cell, masked_embedded_cells], dim=1
        )

        hidden = self.reduce_sequence(concatenated)

        # ================ FC Stack ================
        # logger.debug('  flatten hidden: {0}'.format(hidden))
        hidden = self.fc_stack(hidden)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return H3EmbedConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([H3_INPUT_SIZE])

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape


@DeveloperAPI
@register_encoder("weighted_sum", H3)
class H3WeightedSum(Encoder):
    def __init__(
        self,
        embedding_size: int = 10,
        embeddings_on_cpu: bool = False,
        should_softmax: bool = False,
        fc_layers: Optional[List] = None,
        num_fc_layers: int = 0,
        output_size: int = 10,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Dict = None,
        activation: str = "relu",
        dropout: float = 0,
        encoder_config=None,
        **kwargs,
    ):
        """
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for
               `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memory and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.should_softmax = should_softmax
        self.sum_sequence_reducer = SequenceReducer(reduce_mode="sum")

        self.h3_embed = H3Embed(
            embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            weights_initializer=weights_initializer,
            bias_initializer=bias_initializer,
            reduce_output="None",
        )

        self.register_buffer(
            "aggregation_weights", torch.Tensor(get_initializer(weights_initializer)([H3_INPUT_SIZE, 1]))
        )

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=self.h3_embed.output_shape[0],
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        :param inputs: The input vector fed into the encoder.
               Shape: [batch x H3_INPUT_SIZE], type torch.int8
        :type inputs: Tensor
        """
        # ================ Embeddings ================
        input_vector = inputs
        embedded_h3 = self.h3_embed(input_vector)

        # ================ Weighted Sum ================
        if self.should_softmax:
            weights = torch.softmax(self.aggregation_weights, dim=None)
        else:
            weights = self.aggregation_weights

        hidden = self.sum_sequence_reducer(embedded_h3["encoder_output"] * weights)

        # ================ FC Stack ================
        # logger.debug('  flatten hidden: {0}'.format(hidden))
        hidden = self.fc_stack(hidden)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return H3WeightedSumConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([H3_INPUT_SIZE])

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape


@DeveloperAPI
@register_encoder("rnn", H3)
class H3RNN(Encoder):
    def __init__(
        self,
        embedding_size: int = 10,
        embeddings_on_cpu: bool = False,
        num_layers: int = 1,
        hidden_size: int = 10,
        cell_type: str = "rnn",
        bidirectional: bool = False,
        activation: str = "tanh",
        recurrent_activation: str = "sigmoid",
        use_bias: bool = True,
        unit_forget_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        recurrent_initializer: str = "orthogonal",
        bias_initializer: str = "zeros",
        dropout: float = 0.0,
        recurrent_dropout: float = 0.0,
        reduce_output: str = "last",
        encoder_config=None,
        **kwargs,
    ):
        """
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for
               `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memory and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param num_layers: the number of stacked recurrent layers.
        :type num_layers: Integer
        :param cell_type: the type of recurrent cell to use.
               Available values are: `rnn`, `lstm`, `lstm_block`, `lstm`,
               `ln`, `lstm_cudnn`, `gru`, `gru_block`, `gru_cudnn`.
               For reference about the differences between the cells please
               refer to PyTorch's documentation. We suggest to use the
               `block` variants on CPU and the `cudnn` variants on GPU
               because of their increased speed.
        :type cell_type: str
        :param hidden_size: the size of the state of the rnn.
        :type hidden_size: Integer
        :param bidirectional: if `True` two recurrent networks will perform
               encoding in the forward and backward direction and
               their outputs will be concatenated.
        :type bidirectional: Boolean
        :param activation: Activation function to use.
        :type activation: string
        :param recurrent_activation: Activation function to use for the
                recurrent step.
        :type recurrent_activation: string
        :param use_bias: bool determines where to use a bias vector
        :type use_bias: bool
        :param unit_forget_bias: if True add 1 to the bias forget gate at
               initialization.
        :type unit_forget_bias: bool
        :param weights_initializer: Initializer for the weights (aka kernel)
               matrix
        :type weights_initializer: string
        :param recurrent_initializer: Initializer for the recurrent weights
               matrix
        :type recurrent_initializer: string
        :param bias_initializer: Initializer for the bias vector
        :type bias_initializer: string
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: float
        :param recurrent_dropout: Dropout rate for the RNN encoder of the H3 embeddings.
        :type recurrent_dropout: float
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.embedding_size = embedding_size

        self.h3_embed = H3Embed(
            embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            weights_initializer=weights_initializer,
            bias_initializer=bias_initializer,
            reduce_output="None",
        )

        logger.debug("  RecurrentStack")
        self.recurrent_stack = RecurrentStack(
            input_size=self.h3_embed.output_shape[0],
            max_sequence_length=H3_INPUT_SIZE,
            hidden_size=hidden_size,
            cell_type=cell_type,
            num_layers=num_layers,
            bidirectional=bidirectional,
            use_bias=use_bias,
            dropout=recurrent_dropout,
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        :param inputs: The input vector fed into the encoder.
               Shape: [batch x H3_INPUT_SIZE], type torch.int8
        :type inputs: Tensor
        """

        # ================ Embeddings ================
        embedded_h3 = self.h3_embed(inputs)

        # ================ RNN ================
        hidden, final_state = self.recurrent_stack(embedded_h3["encoder_output"])

        return {"encoder_output": hidden, "encoder_output_state": final_state}

    @staticmethod
    def get_schema_cls():
        return H3RNNConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([H3_INPUT_SIZE])

    @property
    def output_shape(self) -> torch.Size:
        return self.recurrent_stack.output_shape



from abc import ABC
from typing import Dict

from marshmallow_dataclass import dataclass

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import LOSS, TEST, TRAIN, VALIDATION
from ludwig.modules import metric_modules  # noqa: Needed to ensure that the metric registry is populated.
from ludwig.modules.metric_registry import get_metric_registry
from ludwig.schema import utils as schema_utils
from ludwig.schema.hyperopt.executor import ExecutorConfig, ExecutorDataclassField
from ludwig.schema.hyperopt.search_algorithm import BaseSearchAlgorithmConfig, SearchAlgorithmDataclassField


@DeveloperAPI
@dataclass
class HyperoptConfig(schema_utils.BaseMarshmallowConfig, ABC):
    """Basic hyperopt settings."""

    output_feature: str = "combined"  # TODO: make more restrictive

    goal: str = schema_utils.StringOptions(
        options=["minimize", "maximize"],
        default="minimize",
        allow_none=False,
        description=(
            "Indicates if to minimize or maximize a metric or a loss of any of the output features on any of the "
            "dataset splits. Available values are: minimize (default) or maximize."
        ),
    )

    metric: str = schema_utils.StringOptions(
        options=get_metric_registry().keys(),
        default=LOSS,
        allow_none=False,
        description=(
            "The metric that we want to optimize for. The default one is loss, but depending on the type of the "
            "feature defined in output_feature, different metrics and losses are available. Check the metrics section "
            "of the specific output feature type to figure out what metrics are available to use."
        ),
    )

    split: str = schema_utils.StringOptions(
        options=[TRAIN, VALIDATION, TEST],
        default=VALIDATION,
        allow_none=False,
        description=(
            "The split of data that we want to compute our metric on. By default it is the validation split, but "
            "you have the flexibility to specify also train or test splits."
        ),
    )

    search_alg: BaseSearchAlgorithmConfig = SearchAlgorithmDataclassField(
        description=(
            "Specifies the algorithm to sample the defined parameters space. Candidate algorithms are those "
            "found in Ray Tune's Search Algorithms."
        )
    )

    executor: ExecutorConfig = ExecutorDataclassField(
        description=(
            "specifies how to execute the hyperparameter optimization. The execution could happen locally in a serial "
            "manner or in parallel across multiple workers and with GPUs as well if available. The executor section "
            "includes specification for work scheduling and the number of samples to generate."
        )
    )

    parameters: Dict = schema_utils.Dict(allow_none=False)


@DeveloperAPI
def get_hyperopt_jsonschema():
    props = schema_utils.unload_jsonschema_from_marshmallow_class(HyperoptConfig)["properties"]

    return {
        "type": ["object", "null"],
        "properties": props,
        "title": "hyperopt_options",
        "description": "Settings for hyperopt",
    }


@DeveloperAPI
class HyperoptField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(HyperoptConfig, default_missing=True)

    @staticmethod
    def _jsonschema_type_mapping():
        return get_hyperopt_jsonschema()



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import FILL_WITH_CONST, MISSING_VALUE_STRATEGY_OPTIONS, PREPROCESSING, TIMESERIES
from ludwig.schema import utils as schema_utils
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import register_preprocessor
from ludwig.schema.metadata import FEATURE_METADATA
from ludwig.schema.utils import ludwig_dataclass
from ludwig.utils.tokenizers import tokenizer_registry


@DeveloperAPI
@register_preprocessor(TIMESERIES)
@ludwig_dataclass
class TimeseriesPreprocessingConfig(BasePreprocessingConfig):
    tokenizer: str = schema_utils.StringOptions(
        tokenizer_registry.keys(),
        default="space",
        allow_none=False,
        description="Defines how to map from the raw string content of the dataset column to a sequence of elements.",
        parameter_metadata=FEATURE_METADATA[TIMESERIES][PREPROCESSING]["tokenizer"],
    )

    timeseries_length_limit: int = schema_utils.PositiveInteger(
        default=256,
        allow_none=False,
        description="Defines the maximum length of the timeseries. All timeseries longer than this limit are cut off.",
        parameter_metadata=FEATURE_METADATA[TIMESERIES][PREPROCESSING]["timeseries_length_limit"],
    )

    padding_value: float = schema_utils.NonNegativeFloat(
        default=0.0,
        allow_none=False,
        description="Float value that is used for padding and replacing missing values within a row.",
        parameter_metadata=FEATURE_METADATA[TIMESERIES][PREPROCESSING]["padding_value"],
    )

    padding: str = schema_utils.StringOptions(
        ["left", "right"],
        default="right",
        allow_none=False,
        description="The direction of the padding.",
        parameter_metadata=FEATURE_METADATA[TIMESERIES][PREPROCESSING]["padding"],
    )

    missing_value_strategy: str = schema_utils.StringOptions(
        MISSING_VALUE_STRATEGY_OPTIONS,
        default=FILL_WITH_CONST,
        allow_none=False,
        description=(
            "What strategy to follow when there's a missing value in a column. Currently applies only to a row missing "
            "in its entirety, not invididual elements within the row. For now, `NaN` values within a row are filled "
            "using the `padding_value`."
        ),
        parameter_metadata=FEATURE_METADATA[TIMESERIES][PREPROCESSING]["missing_value_strategy"],
    )

    fill_value: str = schema_utils.String(
        default="",
        allow_none=False,
        description=(
            "The value to replace missing values with in case the `missing_value_strategy` is `fill_with_const`."
        ),
        parameter_metadata=FEATURE_METADATA[TIMESERIES][PREPROCESSING]["fill_value"],
    )

    computed_fill_value: str = schema_utils.String(
        default="",
        allow_none=False,
        description=(
            "The internally computed fill value to replace missing values with in case the "
            "`missing_value_strategy` is `fill_with_mode` or `fill_with_mean`."
        ),
        parameter_metadata=FEATURE_METADATA[TIMESERIES][PREPROCESSING]["computed_fill_value"],
    )



import os
from functools import lru_cache
from typing import Dict

import yaml

from ludwig.api_annotations import PublicAPI
from ludwig.datasets import model_configs


@PublicAPI
def model_configs_for_dataset(dataset_name: str) -> Dict[str, Dict]:
    """Returns a dictionary of built-in model configs for the specified dataset.

    Maps config name to ludwig config dict.
    """
    return _get_model_configs(dataset_name)


@lru_cache(maxsize=3)
def _get_model_configs(dataset_name: str) -> Dict[str, Dict]:
    """Returns all model configs for the specified dataset.

    Model configs are named <dataset_name>_<config_name>.yaml
    """
    import importlib.resources

    config_filenames = [
        f for f in importlib.resources.contents(model_configs) if f.endswith(".yaml") and f.startswith(dataset_name)
    ]
    configs = {}
    for config_filename in config_filenames:
        basename = os.path.splitext(config_filename)[0]
        config_name = basename[len(dataset_name) + 1 :]
        configs[config_name] = _load_model_config(config_filename)
    return configs


def _load_model_config(model_config_filename: str):
    """Loads a model config."""
    model_config_path = os.path.join(os.path.dirname(model_configs.__file__), model_config_filename)
    with open(model_config_path) as f:
        return yaml.safe_load(f)



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import MEAN_SQUARED_ERROR, MODEL_ECD, VECTOR
from ludwig.schema import utils as schema_utils
from ludwig.schema.decoders.base import BaseDecoderConfig
from ludwig.schema.decoders.utils import DecoderDataclassField
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import EncoderDataclassField
from ludwig.schema.features.base import BaseInputFeatureConfig, BaseOutputFeatureConfig
from ludwig.schema.features.loss.loss import BaseLossConfig
from ludwig.schema.features.loss.utils import LossDataclassField
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import PreprocessingDataclassField
from ludwig.schema.features.utils import (
    defaults_config_registry,
    ecd_input_config_registry,
    input_mixin_registry,
    output_config_registry,
    output_mixin_registry,
)
from ludwig.schema.metadata import FEATURE_METADATA
from ludwig.schema.metadata.parameter_metadata import INTERNAL_ONLY
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass


@DeveloperAPI
@input_mixin_registry.register(VECTOR)
@ludwig_dataclass
class VectorInputFeatureConfigMixin(BaseMarshmallowConfig):
    """VectorInputFeatureConfigMixin is a dataclass that configures the parameters used in both the vector input
    feature and the vector global defaults section of the Ludwig Config."""

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type=VECTOR)

    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=VECTOR,
        default="dense",
    )


@DeveloperAPI
@ecd_input_config_registry.register(VECTOR)
@ludwig_dataclass
class VectorInputFeatureConfig(BaseInputFeatureConfig, VectorInputFeatureConfigMixin):
    """VectorInputFeatureConfig is a dataclass that configures the parameters used for a vector input feature."""

    pass


@DeveloperAPI
@output_mixin_registry.register(VECTOR)
@ludwig_dataclass
class VectorOutputFeatureConfigMixin(BaseMarshmallowConfig):
    """VectorOutputFeatureConfigMixin is a dataclass that configures the parameters used in both the vector output
    feature and the vector global defaults section of the Ludwig Config."""

    decoder: BaseDecoderConfig = DecoderDataclassField(
        feature_type=VECTOR,
        default="projector",
    )

    loss: BaseLossConfig = LossDataclassField(
        feature_type=VECTOR,
        default=MEAN_SQUARED_ERROR,
    )


@DeveloperAPI
@output_config_registry.register(VECTOR)
@ludwig_dataclass
class VectorOutputFeatureConfig(BaseOutputFeatureConfig, VectorOutputFeatureConfigMixin):
    """VectorOutputFeatureConfig is a dataclass that configures the parameters used for a vector output feature."""

    dependencies: list = schema_utils.List(
        default=[],
        description="List of input features that this feature depends on.",
        parameter_metadata=FEATURE_METADATA[VECTOR]["dependencies"],
    )

    default_validation_metric: str = schema_utils.StringOptions(
        [MEAN_SQUARED_ERROR],
        default=MEAN_SQUARED_ERROR,
        description="Internal only use parameter: default validation metric for binary output feature.",
        parameter_metadata=INTERNAL_ONLY,
    )

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type="vector_output")

    reduce_dependencies: str = schema_utils.ReductionOptions(
        default=None,
        description="How to reduce the dependencies of the output feature.",
        parameter_metadata=FEATURE_METADATA[VECTOR]["reduce_dependencies"],
    )

    reduce_input: str = schema_utils.ReductionOptions(
        default=None,
        description="How to reduce an input that is not a vector, but a matrix or a higher order tensor, on the first "
        "dimension (second if you count the batch dimension)",
        parameter_metadata=FEATURE_METADATA[VECTOR]["reduce_input"],
    )

    softmax: bool = schema_utils.Boolean(
        default=False,
        description="Determines whether to apply a softmax at the end of the decoder. This is useful for predicting a "
        "vector of values that sum up to 1 and can be interpreted as probabilities.",
        parameter_metadata=FEATURE_METADATA[VECTOR]["softmax"],
    )

    vector_size: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The size of the vector. If None, the vector size will be inferred from the data.",
        parameter_metadata=FEATURE_METADATA[VECTOR]["vector_size"],
    )


@DeveloperAPI
@defaults_config_registry.register(VECTOR)
@ludwig_dataclass
class VectorDefaultsConfig(VectorInputFeatureConfigMixin, VectorOutputFeatureConfigMixin):
    pass



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import contextlib
from abc import ABC, abstractmethod

from ludwig.distributed import DistributedStrategy
from ludwig.utils.defaults import default_random_seed


class Dataset(ABC):
    @abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @contextlib.contextmanager
    @abstractmethod
    def initialize_batcher(
        self,
        batch_size: int = 128,
        should_shuffle: bool = True,
        random_seed: int = default_random_seed,
        ignore_last: bool = False,
        distributed: DistributedStrategy = None,
    ):
        raise NotImplementedError()

    def to_df(self):
        raise NotImplementedError()

    @property
    def in_memory_size_bytes(self):
        raise NotImplementedError()


class DatasetManager(ABC):
    @abstractmethod
    def create(self, dataset, config, training_set_metadata):
        raise NotImplementedError()

    @abstractmethod
    def save(self, cache_path, dataset, config, training_set_metadata, tag):
        raise NotImplementedError()

    @abstractmethod
    def can_cache(self, skip_save_processed_input):
        raise NotImplementedError()

    @property
    @abstractmethod
    def data_format(self):
        raise NotImplementedError()



"""Uses heuristics to build ludwig configuration file:

(1) infer types based on dataset
(2) populate with
    - default combiner parameters,
    - preprocessing parameters,
    - combiner specific default training parameters,
    - combiner specific hyperopt space
    - feature parameters
(3) add machineresources
    (base implementation -- # CPU, # GPU)
"""

import os
from dataclasses import dataclass
from typing import Any, Dict, List, Set, Union

import dask.dataframe as dd
import numpy as np
import pandas as pd
import yaml
from dataclasses_json import dataclass_json, LetterCase

from ludwig.api_annotations import DeveloperAPI
from ludwig.backend import Backend
from ludwig.constants import (
    COLUMN,
    COMBINER,
    ENCODER,
    EXECUTOR,
    HYPEROPT,
    INPUT_FEATURES,
    PREPROCESSING,
    SCHEDULER,
    SEARCH_ALG,
    SPLIT,
    TEXT,
    TYPE,
)
from ludwig.profiling import dataset_profile_pb2
from ludwig.profiling.dataset_profile import get_dataset_profile_proto, get_dataset_profile_view
from ludwig.types import ModelConfigDict
from ludwig.utils.automl.data_source import DataSource, wrap_data_source
from ludwig.utils.automl.field_info import FieldConfig, FieldInfo, FieldMetadata
from ludwig.utils.automl.type_inference import infer_type, should_exclude
from ludwig.utils.data_utils import load_yaml
from ludwig.utils.misc_utils import merge_dict
from ludwig.utils.system_utils import Resources

PATH_HERE = os.path.abspath(os.path.dirname(__file__))
CONFIG_DIR = os.path.join(PATH_HERE, "defaults")

BASE_AUTOML_CONFIG = os.path.join(CONFIG_DIR, "base_automl_config.yaml")
REFERENCE_CONFIGS = os.path.join(CONFIG_DIR, "reference_configs.yaml")

combiner_defaults = {
    "concat": os.path.join(CONFIG_DIR, "combiner/concat_config.yaml"),
    "tabnet": os.path.join(CONFIG_DIR, "combiner/tabnet_config.yaml"),
    "transformer": os.path.join(CONFIG_DIR, "combiner/transformer_config.yaml"),
}

encoder_defaults = {"text": {"bert": os.path.join(CONFIG_DIR, "text/bert_config.yaml")}}

# Cap for number of distinct values to return.
MAX_DISTINCT_VALUES_TO_RETURN = 10


@DeveloperAPI
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DatasetInfo:
    fields: List[FieldInfo]
    row_count: int
    size_bytes: int = -1


def allocate_experiment_resources(resources: Resources) -> dict:
    """Allocates ray trial resources based on available resources.

    # Inputs
    :param resources (dict) specifies all available GPUs, CPUs and associated
        metadata of the machines (i.e. memory)

    # Return
    :return: (dict) gpu and cpu resources per trial
    """
    # TODO (ASN):
    # (1) expand logic to support multiple GPUs per trial (multi-gpu training)
    # (2) add support for kubernetes namespace (if applicable)
    # (3) add support for smarter allocation based on size of GPU memory
    experiment_resources = {"cpu_resources_per_trial": 1}
    gpu_count, cpu_count = resources.gpus, resources.cpus
    if gpu_count > 0:
        experiment_resources.update({"gpu_resources_per_trial": 1})
        if cpu_count > 1:
            cpus_per_trial = max(int(cpu_count / gpu_count), 1)
            experiment_resources["cpu_resources_per_trial"] = cpus_per_trial

    return experiment_resources


def get_resource_aware_hyperopt_config(
    experiment_resources: Dict[str, Any], time_limit_s: Union[int, float], random_seed: int
) -> Dict[str, Any]:
    """Returns a Ludwig config with the hyperopt section populated with appropriate parameters.

    Hyperopt parameters are intended to be appropriate for the given resources and time limit.
    """
    executor = experiment_resources
    executor.update({"time_budget_s": time_limit_s})
    if time_limit_s is not None:
        executor.update({SCHEDULER: {"max_t": time_limit_s}})

    return {
        HYPEROPT: {
            SEARCH_ALG: {"random_state_seed": random_seed},
            EXECUTOR: executor,
        },
    }


def _get_stratify_split_config(field_meta: FieldMetadata) -> dict:
    return {
        PREPROCESSING: {
            SPLIT: {
                TYPE: "stratify",
                COLUMN: field_meta.name,
            }
        }
    }


def get_default_automl_hyperopt() -> Dict[str, Any]:
    """Returns general, default settings for hyperopt.

    For example:
    - We set a random_state_seed for sample sequence repeatability
    - We use an increased reduction_factor to get more pruning/exploration.

    TODO: If settings seem reasonable, consider building this into the hyperopt schema, directly.
    """
    return yaml.safe_load(
        """
  search_alg:
    type: hyperopt
  executor:
    type: ray
    num_samples: 10
    time_budget_s: 3600
    scheduler:
      type: async_hyperband
      time_attr: time_total_s
      max_t: 3600
      grace_period: 72
      reduction_factor: 5
"""
    )


def create_default_config(
    features_config: ModelConfigDict,
    dataset_info: DatasetInfo,
    target_name: Union[str, List[str]],
    time_limit_s: Union[int, float],
    random_seed: int,
    imbalance_threshold: float = 0.9,
    backend: Backend = None,
) -> dict:
    """Returns auto_train configs for three available combiner models. Coordinates the following tasks:

    - extracts fields and generates list of FieldInfo objects
    - gets field metadata (i.e avg. words, total non-null entries)
    - builds input_features and output_features section of config
    - for imbalanced datasets, a preprocessing section is added to perform stratified sampling if the imbalance ratio
      is smaller than imbalance_threshold
    - for each combiner, adds default training, hyperopt
    - infers resource constraints and adds gpu and cpu resource allocation per
      trial

    # Inputs
    :param dataset_info: (str) filepath Dataset Info object.
    :param target_name: (str, List[str]) name of target feature
    :param time_limit_s: (int, float) total time allocated to auto_train. acts
                                    as the stopping parameter
    :param random_seed: (int, default: `42`) a random seed that will be used anywhere
                        there is a call to a random number generator, including
                        hyperparameter search sampling, as well as data splitting,
                        parameter initialization and training set shuffling
    :param imbalance_threshold: (float) maximum imbalance ratio (minority / majority) to perform stratified sampling
    :param backend: (Backend) backend to use for training.

    # Return
    :return: (dict) dictionaries contain auto train config files for all available
    combiner types
    """
    base_automl_config = load_yaml(BASE_AUTOML_CONFIG)
    base_automl_config.update(features_config)

    targets = convert_targets(target_name)
    features_metadata = get_field_metadata(dataset_info.fields, dataset_info.row_count, targets)

    # Handle expensive features for CPU
    resources = backend.get_available_resources()
    for ifeature in base_automl_config[INPUT_FEATURES]:
        if resources.gpus == 0:
            if ifeature[TYPE] == TEXT:
                # When no GPUs are available, default to the embed encoder, which is fast enough for CPU
                ifeature[ENCODER] = {"type": "embed"}

    # create set of all feature types appearing in the dataset
    feature_types = [[feat[TYPE] for feat in features] for features in features_config.values()]
    feature_types = set(sum(feature_types, []))

    model_configs = {}

    # update hyperopt config
    experiment_resources = allocate_experiment_resources(resources)
    base_automl_config = merge_dict(
        base_automl_config, get_resource_aware_hyperopt_config(experiment_resources, time_limit_s, random_seed)
    )

    # add preprocessing section if single output feature is imbalanced
    outputs_metadata = [f for f in features_metadata if f.mode == "output"]
    if len(outputs_metadata) == 1:
        of_meta = outputs_metadata[0]
        is_categorical = of_meta.config.type in ["category", "binary"]
        is_imbalanced = of_meta.imbalance_ratio < imbalance_threshold
        if is_categorical and is_imbalanced:
            base_automl_config.update(_get_stratify_split_config(of_meta))

    model_configs["base_config"] = base_automl_config

    # read in all encoder configs
    for feat_type, default_configs in encoder_defaults.items():
        if feat_type in feature_types:
            if feat_type not in model_configs.keys():
                model_configs[feat_type] = {}
            for encoder_name, encoder_config_path in default_configs.items():
                model_configs[feat_type][encoder_name] = load_yaml(encoder_config_path)

    # read in all combiner configs
    model_configs[COMBINER] = {}
    for combiner_type, default_config in combiner_defaults.items():
        combiner_config = load_yaml(default_config)
        model_configs[COMBINER][combiner_type] = combiner_config

    return model_configs


# Read in the score and configuration of a reference model trained by Ludwig for each dataset in a list.
def get_reference_configs() -> dict:
    reference_configs = load_yaml(REFERENCE_CONFIGS)
    return reference_configs


def get_dataset_info(df: Union[pd.DataFrame, dd.core.DataFrame]) -> DatasetInfo:
    """Constructs FieldInfo objects for each feature in dataset. These objects are used for downstream type
    inference.

    # Inputs
    :param df: (Union[pd.DataFrame, dd.core.DataFrame]) Pandas or Dask dataframe.

    # Return
    :return: (DatasetInfo) Structure containing list of FieldInfo objects.
    """
    source = wrap_data_source(df)
    return get_dataset_info_from_source(source)


def is_field_boolean(source: DataSource, field: str) -> bool:
    """Returns a boolean indicating whether the object field should have a bool dtype.

    Columns with object dtype that have 3 distinct values of which one is Nan/None is a bool type column.
    """
    unique_values = source.df[field].unique()
    if len(unique_values) <= 3:
        for entry in unique_values:
            try:
                if np.isnan(entry):
                    continue
            except TypeError:
                # For some field types such as object arrays, np.isnan throws a TypeError
                # In this case, do nothing and proceed to checking if the entry is a bool object
                pass
            if isinstance(entry, bool):
                continue
            return False
        return True
    return False


@DeveloperAPI
def get_dataset_profile_from_source(source: DataSource) -> dataset_profile_pb2.DatasetProfile:
    return get_dataset_profile_proto(get_dataset_profile_view(source.df))


@DeveloperAPI
def get_dataset_info_from_source(source: DataSource) -> DatasetInfo:
    """Constructs FieldInfo objects for each feature in dataset. These objects are used for downstream type
    inference.

    # Inputs
    :param source: (DataSource) A wrapper around a data source, which may represent a pandas or Dask dataframe.

    # Return
    :return: (DatasetInfo) Structure containing list of FieldInfo objects.
    """
    row_count = len(source)
    fields = []
    for field in source.columns:
        dtype = source.get_dtype(field)
        num_distinct_values, distinct_values, distinct_values_balance = source.get_distinct_values(
            field, MAX_DISTINCT_VALUES_TO_RETURN
        )
        nonnull_values = source.get_nonnull_values(field)
        image_values = source.get_image_values(field)
        audio_values = source.get_audio_values(field)
        avg_words = None
        if dtype == "object":
            # Check if it is a nullboolean field. We do this since if you read a csv with
            # pandas that has a column of booleans and some missing values, the column is
            # interpreted as object dtype instead of bool
            if is_field_boolean(source, field):
                dtype = "bool"
        if source.is_string_type(dtype):
            avg_words = source.get_avg_num_tokens(field)
        fields.append(
            FieldInfo(
                name=field,
                dtype=dtype,
                distinct_values=distinct_values,
                num_distinct_values=num_distinct_values,
                distinct_values_balance=distinct_values_balance,
                nonnull_values=nonnull_values,
                image_values=image_values,
                audio_values=audio_values,
                avg_words=avg_words,
            )
        )
    return DatasetInfo(fields=fields, row_count=row_count, size_bytes=source.size_bytes())


def get_features_config(
    fields: List[FieldInfo],
    row_count: int,
    target_name: Union[str, List[str]] = None,
) -> dict:
    """Constructs FieldInfo objects for each feature in dataset. These objects are used for downstream type
    inference.

    # Inputs
    :param fields: (List[FieldInfo]) FieldInfo objects for all fields in dataset
    :param row_count: (int) total number of entries in original dataset
    :param target_name (str, List[str]) name of target feature

    # Return
    :return: (dict) section of auto_train config for input_features and output_features
    """
    targets = convert_targets(target_name)
    metadata = get_field_metadata(fields, row_count, targets)
    return get_config_from_metadata(metadata, targets)


def convert_targets(target_name: Union[str, List[str]] = None) -> Set[str]:
    targets = target_name
    if isinstance(targets, str):
        targets = [targets]
    if targets is None:
        targets = []
    return set(targets)


def get_config_from_metadata(metadata: List[FieldMetadata], targets: Set[str] = None) -> dict:
    """Builds input/output feature sections of auto-train config using field metadata.

    # Inputs
    :param metadata: (List[FieldMetadata]) field descriptions
    :param targets (Set[str]) names of target features

    # Return
    :return: (dict) section of auto_train config for input_features and output_features
    """
    config = {
        "input_features": [],
        "output_features": [],
    }

    for field_meta in metadata:
        if field_meta.name in targets:
            config["output_features"].append(field_meta.config.to_dict())
        elif not field_meta.excluded and field_meta.mode == "input":
            config["input_features"].append(field_meta.config.to_dict())

    return config


@DeveloperAPI
def get_field_metadata(fields: List[FieldInfo], row_count: int, targets: Set[str] = None) -> List[FieldMetadata]:
    """Computes metadata for each field in dataset.

    # Inputs
    :param fields: (List[FieldInfo]) FieldInfo objects for all fields in dataset
    :param row_count: (int) total number of entries in original dataset
    :param targets (Set[str]) names of target features

    # Return
    :return: (List[FieldMetadata]) list of objects containing metadata for each field
    """

    metadata = []
    column_count = len(fields)
    for idx, field in enumerate(fields):
        missing_value_percent = 1 - float(field.nonnull_values) / row_count
        dtype = infer_type(field, missing_value_percent, row_count)
        metadata.append(
            FieldMetadata(
                name=field.name,
                config=FieldConfig(
                    name=field.name,
                    column=field.name,
                    type=dtype,
                ),
                excluded=should_exclude(idx, field, dtype, column_count, row_count, targets),
                mode=infer_mode(field, targets),
                missing_values=missing_value_percent,
                imbalance_ratio=field.distinct_values_balance,
            )
        )

    return metadata


def infer_mode(field: FieldInfo, targets: Set[str] = None) -> str:
    if field.name in targets:
        return "output"
    if field.name.lower() == "split":
        return "split"
    return "input"



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import logging
import os
import sys
from typing import Optional

from ludwig.api import LudwigModel
from ludwig.contrib import add_contrib_callback_args
from ludwig.globals import LUDWIG_VERSION
from ludwig.utils.neuropod_utils import export_neuropod as utils_export_neuropod
from ludwig.utils.print_utils import get_logging_level_registry, print_ludwig
from ludwig.utils.triton_utils import export_triton as utils_export_triton

logger = logging.getLogger(__name__)


def export_torchscript(
    model_path: str, model_only: bool = False, output_path: Optional[str] = None, device: Optional[str] = None, **kwargs
) -> None:
    """Exports a model to torchscript.

    # Inputs

    :param model_path: (str) filepath to pre-trained model.
    :param model_only: (bool, default: `False`) If true, scripts and exports the model only.
    :param output_path: directory to store torchscript. If `None`, defaults to model_path

    # Return
    :returns: (`None`)
    """
    logger.info(f"Model path: {model_path}")
    logger.info(f"Saving model only: {model_only}")

    if output_path is None:
        logger.info("output_path is None, defaulting to model_path")
        output_path = model_path
    logger.info(f"Output path: {output_path}")
    logger.info("\n")

    model = LudwigModel.load(model_path)
    os.makedirs(output_path, exist_ok=True)
    model.save_torchscript(output_path, model_only=model_only, device=device)

    logger.info(f"Saved to: {output_path}")


def export_triton(model_path, output_path="model_repository", model_name="ludwig_model", model_version=1, **kwargs):
    """Exports a model in torchscript format with config for Triton serving.

    # Inputs

    :param model_path: (str) filepath to pre-trained model.
    :param output_path: (str, default: `'model_repository'`)  directory to store the
        triton models.
    :param model_name: (str, default: `'ludwig_model'`) save triton under this name.
    :param model_name: (int, default: `1`) save neuropod under this verison.

    # Return

    :returns: (`None`)
    """
    logger.info(f"Model path: {model_path}")
    logger.info(f"Output path: {output_path}")
    logger.info(f"Model name: {model_name}")
    logger.info(f"Model version: {model_version}")
    logger.info("\n")

    model = LudwigModel.load(model_path)
    os.makedirs(output_path, exist_ok=True)

    utils_export_triton(model=model, output_path=output_path, model_name=model_name, model_version=model_version)

    logger.info(f"Saved to: {output_path}")


def export_neuropod(model_path, output_path="neuropod", model_name="neuropod", **kwargs):
    """Exports a model to Neuropod.

    # Inputs

    :param model_path: (str) filepath to pre-trained model.
    :param output_path: (str, default: `'neuropod'`)  directory to store the
        neuropod model.
    :param model_name: (str, default: `'neuropod'`) save neuropod under this
        name.

    # Return

    :returns: (`None`)
    """
    logger.info(f"Model path: {model_path}")
    logger.info(f"Output path: {output_path}")
    logger.info("\n")

    model = LudwigModel.load(model_path)
    os.makedirs(output_path, exist_ok=True)
    utils_export_neuropod(model, output_path, model_name)

    logger.info(f"Saved to: {output_path}")


def export_mlflow(model_path, output_path="mlflow", registered_model_name=None, **kwargs):
    """Exports a model to MLflow.

    # Inputs

    :param model_path: (str) filepath to pre-trained model.
    :param output_path: (str, default: `'mlflow'`)  directory to store the
        mlflow model.
    :param registered_model_name: (str, default: `None`) save mlflow under this
        name in the model registry. Saved locally if `None`.

    # Return

    :returns: (`None`)
    """
    logger.info(f"Model path: {model_path}")
    logger.info(f"Output path: {output_path}")
    logger.info("\n")

    from ludwig.contribs.mlflow.model import export_model

    export_model(model_path, output_path, registered_model_name)

    logger.info(f"Saved to: {output_path}")


def cli_export_torchscript(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model " "and saves it as torchscript.",
        prog="ludwig export_torchscript",
        usage="%(prog)s [options]",
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)
    parser.add_argument(
        "-mo",
        "--model_only",
        help="Script and export the model only.",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--device",
        type=str,
        help=(
            'Device to use for torchscript tracing (e.g. "cuda" or "cpu"). Ideally, this is the same as the device '
            "used when the model is loaded."
        ),
        default=None,
    )

    # -----------------
    # Output parameters
    # -----------------
    parser.add_argument(
        "-op",
        "--output_path",
        type=str,
        help="path where to save the export model. If not specified, defaults to model_path.",
        default=None,
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("export_torchscript", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.export")

    print_ludwig("Export Torchscript", LUDWIG_VERSION)

    export_torchscript(**vars(args))


def cli_export_triton(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model " "and saves it as torchscript for Triton.",
        prog="ludwig export_neuropod",
        usage="%(prog)s [options]",
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)
    parser.add_argument("-mn", "--model_name", help="model name", default="ludwig_model")
    parser.add_argument("-mv", "--model_version", type=int, help="model version", default=1)

    # -----------------
    # Output parameters
    # -----------------
    parser.add_argument("-op", "--output_path", type=str, help="path where to save the export model", required=True)

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("export_triton", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.export")

    print_ludwig("Export Triton", LUDWIG_VERSION)

    export_triton(**vars(args))


def cli_export_neuropod(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model " "and saves it as a Neuropod.",
        prog="ludwig export_neuropod",
        usage="%(prog)s [options]",
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)
    parser.add_argument("-mn", "--model_name", help="model name", default="neuropod")

    # -----------------
    # Output parameters
    # -----------------
    parser.add_argument("-op", "--output_path", type=str, help="path where to save the export model", required=True)

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("export_neuropod", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.export")

    print_ludwig("Export Neuropod", LUDWIG_VERSION)

    export_neuropod(**vars(args))


def cli_export_mlflow(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model " "and saves it as an MLFlow model.",
        prog="ludwig export_mlflow",
        usage="%(prog)s [options]",
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)
    parser.add_argument(
        "-mn",
        "--registered_model_name",
        help="model name to upload to in MLflow model registry",
    )

    # -----------------
    # Output parameters
    # -----------------
    parser.add_argument(
        "-op", "--output_path", type=str, help="path where to save the exported model", default="mlflow"
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("export_mlflow", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.export")

    print_ludwig("Export MLFlow", LUDWIG_VERSION)

    export_mlflow(**vars(args))


if __name__ == "__main__":
    if len(sys.argv) > 1:
        if sys.argv[1] == "savedmodel":
            cli_export_torchscript(sys.argv[2:])
        elif sys.argv[1] == "mlflow":
            cli_export_mlflow(sys.argv[2:])
        elif sys.argv[1] == "triton":
            cli_export_triton(sys.argv[2:])
        elif sys.argv[1] == "neuropod":
            cli_export_neuropod(sys.argv[2:])
        else:
            print("Unrecognized command")
    else:
        print("Unrecognized command")



"""Utility functions related to sequence decoders."""

from typing import Dict, Tuple

import torch

from ludwig.constants import ENCODER_OUTPUT_STATE, HIDDEN
from ludwig.modules.reduction_modules import SequenceReducer


def repeat_2D_tensor(tensor, k):
    """Repeats a 2D-tensor k times over the first dimension.

    For example:
    Input: Tensor of [batch_size, state_size], k=2
    Output: Tensor of [k, batch_size, state_size]
    """
    if len(tensor.size()) > 2:
        raise ValueError("Cannot repeat a non-2D tensor with this method.")
    return tensor.repeat(k, 1, 1)


def get_rnn_init_state(
    combiner_outputs: Dict[str, torch.Tensor], sequence_reducer: SequenceReducer, num_layers: int
) -> torch.Tensor:
    """Computes the hidden state that the RNN decoder should start with.

    Args:
        combiner_outputs: Dictionary of tensors from the outputs of the combiner and other output features.
        sequence_reducer: SequenceReducer to reduce rank-3 to rank-2.
        num_layers: Number of layers the decoder uses.

    Returns:
        Tensor of [num_layers, batch_size, hidden_size].
    """
    if ENCODER_OUTPUT_STATE not in combiner_outputs:
        # Use the combiner's hidden state.
        encoder_output_state = combiner_outputs[HIDDEN]
    else:
        # Use the encoder's output state.
        encoder_output_state = combiner_outputs[ENCODER_OUTPUT_STATE]
        if isinstance(encoder_output_state, tuple):
            if len(encoder_output_state) == 2:
                # LSTM encoder. Use the hidden state and ignore the cell state.
                encoder_output_state = encoder_output_state[0]
            elif len(encoder_output_state) == 4:
                # Bi-directional LSTM encoder. Use the average of hidden states and ignore cell state.
                encoder_output_state = torch.mean([encoder_output_state[0], encoder_output_state[2]])
            else:
                raise ValueError(
                    f"Invalid sequence decoder inputs with keys: {combiner_outputs.keys()} with extracted encoder "
                    + f"state: {encoder_output_state.size()} that was invalid. Please double check the compatibility "
                    + "of your encoder and decoder."
                )

    if len(encoder_output_state.size()) > 3:
        raise ValueError("Init state for RNN decoders only works for 1d or 2d tensors (encoder_output).")

    if len(encoder_output_state.size()) == 3:
        # Reduce to [batch_size, hidden_size].
        encoder_output_state = sequence_reducer(encoder_output_state)

    return repeat_2D_tensor(encoder_output_state, num_layers)


def get_lstm_init_state(
    combiner_outputs: Dict[str, torch.Tensor], sequence_reducer: SequenceReducer, num_layers: int
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Returns the states that the LSTM decoder should start with.

    Args:
        combiner_outputs: Dictionary of tensors from the outputs of the combiner and other output features.
        sequence_reducer: SequenceReducer to reduce rank-3 to rank-2.
        num_layers: Number of layers the decoder uses.

    Returns:
        Tuple of 2 tensors (decoder hidden state, decoder cell state), each [num_layers, batch_size, hidden_size].
    """
    if "encoder_output_state" not in combiner_outputs:
        # Use the combiner's hidden state.
        decoder_hidden_state = combiner_outputs[HIDDEN]
        decoder_cell_state = torch.clone(decoder_hidden_state)
    else:
        # Use the encoder's output state.
        encoder_output_state = combiner_outputs[ENCODER_OUTPUT_STATE]
        if not isinstance(encoder_output_state, tuple):
            decoder_hidden_state = encoder_output_state
            decoder_cell_state = decoder_hidden_state
        else:
            if len(encoder_output_state) == 2:
                # The encoder was probably an LSTM.
                decoder_hidden_state, decoder_cell_state = encoder_output_state
            elif len(encoder_output_state) == 4:
                # The encoder was probably a bi-LSTM.
                # Use the average of the encoder's hidden states for hidden state.
                # Use the average of the encoder's cell states for cell state.
                decoder_hidden_state = torch.mean([encoder_output_state[0], encoder_output_state[2]])
                decoder_cell_state = torch.mean([encoder_output_state[1], encoder_output_state[3]])
            else:
                raise ValueError(
                    f"Invalid sequence decoder inputs with keys: {combiner_outputs.keys()} with extracted encoder "
                    + f"state: {encoder_output_state} that was invalid. Please double check the compatibility of your "
                    + "encoder and decoder."
                )

    # Check rank and reduce if necessary.
    if len(decoder_hidden_state.size()) > 3 or len(decoder_cell_state.size()) > 3:
        raise ValueError(
            f"Invalid sequence decoder inputs with keys: {combiner_outputs.keys()} with extracted encoder "
            + f"state: {decoder_hidden_state.size()} that was invalid. Please double check the compatibility "
            + "of your encoder and decoder."
        )
    if len(decoder_hidden_state.size()) == 3:
        decoder_hidden_state = sequence_reducer(decoder_hidden_state)
    if len(decoder_cell_state.size()) == 3:
        decoder_cell_state = sequence_reducer(decoder_cell_state)

    # Repeat over the number of layers.
    return repeat_2D_tensor(decoder_hidden_state, num_layers), repeat_2D_tensor(decoder_cell_state, num_layers)



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import time
from typing import Type, Union

import psutil
import torch

from ludwig.backend.base import Backend, LocalPreprocessingMixin
from ludwig.data.dataset.pandas import PandasDatasetManager
from ludwig.distributed.horovod import HorovodStrategy
from ludwig.models.base import BaseModel
from ludwig.models.predictor import Predictor
from ludwig.trainers.trainer import Trainer
from ludwig.types import HyperoptConfigDict
from ludwig.utils.batch_size_tuner import BatchSizeEvaluator
from ludwig.utils.system_utils import Resources
from ludwig.utils.torch_utils import initialize_pytorch


class HorovodBackend(LocalPreprocessingMixin, Backend):
    BACKEND_TYPE = "horovod"

    def __init__(self, **kwargs):
        super().__init__(dataset_manager=PandasDatasetManager(self), **kwargs)
        self._distributed = None

    def initialize(self):
        self._distributed = HorovodStrategy()

    def initialize_pytorch(self, *args, **kwargs):
        initialize_pytorch(
            *args, local_rank=self._distributed.local_rank(), local_size=self._distributed.local_size(), **kwargs
        )

    def create_trainer(self, **kwargs) -> "BaseTrainer":  # noqa: F821
        return Trainer(distributed=self._distributed, **kwargs)

    def create_predictor(self, model: BaseModel, **kwargs):
        return Predictor(model, distributed=self._distributed, **kwargs)

    def sync_model(self, model):
        # Model weights are only saved on the coordinator, so broadcast
        # to all other ranks
        self._distributed.sync_model(model)

    def broadcast_return(self, fn):
        """Returns the result of calling `fn` on coordinator, broadcast to all other ranks.

        Specifically, `fn` is only executed on coordinator, but its result is returned by every rank by broadcasting the
        return value from coordinator.
        """
        result = fn() if self.is_coordinator() else None
        if self._distributed:
            name = f"broadcast_return_{int(time.time())}"
            result = self._distributed.broadcast_object(result, name=name)
        return result

    def is_coordinator(self):
        return self._distributed.rank() == 0

    @property
    def num_nodes(self) -> int:
        return self._distributed.size()

    def get_available_resources(self) -> Resources:
        # TODO(travis): this double-counts on the same device, it should use a cross-communicator instead
        cpus = torch.as_tensor([psutil.cpu_count()], dtype=torch.int)
        cpus = self._distributed.allreduce(cpus).item()

        gpus = torch.as_tensor([torch.cuda.device_count()], dtype=torch.int)
        gpus = self._distributed.allreduce(gpus).item()

        return Resources(cpus=cpus, gpus=gpus)

    def max_concurrent_trials(self, hyperopt_config: HyperoptConfigDict) -> Union[int, None]:
        # Return None since there is no Ray component
        return None

    def tune_batch_size(self, evaluator_cls: Type[BatchSizeEvaluator], dataset_len: int) -> int:
        evaluator = evaluator_cls()
        return evaluator.select_best_batch_size(dataset_len)



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Optional

import pandas as pd

from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class HiggsLoader(DatasetLoader):
    def __init__(self, config: DatasetConfig, cache_dir: Optional[str] = None, add_validation_set=True):
        super().__init__(config, cache_dir)
        self.add_validation_set = add_validation_set

    def load_file_to_dataframe(self, file_path: str) -> pd.DataFrame:
        """Loads a file into a dataframe."""
        return pd.read_csv(file_path, header=None)

    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        processed_df = super().transform_dataframe(dataframe)
        if self.add_validation_set:
            processed_df["split"] = [0] * 10000000 + [1] * 500000 + [2] * 500000
        else:
            processed_df["split"] = [0] * 10500000 + [2] * 500000
        return processed_df
